package Authenticate;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	CheckAdminUser
	CheckAuthentication
	LoadUsers
	SaveUsers
);
use strict;
use warnings;
use Common;
use Multiuser;
use Data;
use HTML;

# ---------------------------------------------------------------------------------------------
# variables
use vars qw ($LastToken);
$LastToken = 0;
use vars qw ($passwordFile);
$passwordFile = "${DataDir}/$Program.pw";
$Backups{$passwordFile} = TRUE;

# ---------------------------------------------------------------------------------------------
sub GetToken {
	my ($username) = @_;
	$LastToken++;
	$Tokens{$LastToken}{lastaccess} = $Now{seconds};
	$Tokens{$LastToken}{username}   = $username;
	StartSession($LastToken);
	InitialRecordFields();
	return $LastToken;
}

# ---------------------------------------------------------------------------------------------
sub ShowLoginForm {
	my ($message) = @_;
	HTTPHeader();
	HTMLHeader();
	print HTMLStart("body style='background-color:lightgray;' onload=\"javascript:document.getElementById('username').focus();\"");
	print HTMLStart('center');
	print HTMLStartEnd( "h1 style='background-color:pink;width:50%;'", "$Title <small><small>Version $Version</small></small><hr class='nomargin'>" );
	print HTMLStart("div");
	print HTMLStart('table border');
	print HTMLStart("form method=POST action='login' ");
	print HTMLStartEnd( "tr style='background-color:lightblue;'", HTMLEnclosed( "th", 'Username' ), HTMLEnclosed( "th", 'Password' ), HTMLEnclosed( "th", 'New Password') );
	print HTMLStart('tr');
	print HTMLStartEnd( 'th', HTMLSingle("input id='username' type=text name='username'"));
	print HTMLStartEnd( 'th', HTMLSingle("input type=password name='password'"));
	print HTMLStartEnd( 'th', HTMLSingle("input type=text name='newpassword'"));
	print HTMLEnd('tr');
	print HTMLStart('tr');
	print HTMLStartEnd( 'th', );
	print HTMLStartEnd( 'th', HTMLEnclosed("input type=submit name='login'  value='Login'") );
	print HTMLStartEnd( 'th', HTMLEnclosed("input type=submit name='setpass' value='Change Password'") );
	print HTMLEnd('tr');
	print HTMLStartEnd( 'tr', HTMLEnclosed( 'th colspan=3', HTMLEnclosed("input type=submit name='cancel' value='Cancel'") ) );
	print HTMLStartEnd( "tr style='background-color:red;'", HTMLEnclosed( "th colspan=3", $message ) ) if $message;
	print HTMLEnd('form');
	print HTMLEnd('table');
	print HTMLEnd('div');
	print HTMLEnd('center');
	print HTMLEnd('body');
	print HTMLEnd('html');
	return FALSE;
}

# ---------------------------------------------------------------------------------------------
sub PasswordHash {
	my ($password) = @_;
	return $password if !$password;
	my $hash = '';
	for ( my $charnum = 0; $charnum < length($password); $charnum++ ) {
		my $char  = substr( $password, $charnum, 1 );
		my $ascii = ord($char);
		my $shift = $charnum * 7;
		$ascii += 256 if $ascii < $shift;
		$ascii -= $shift;
		$hash .= sprintf "%.3d", $ascii;
	}
	return $hash;
}

sub LoadUsers {
	return if !$properties{authentication};
	%Users = ();
	open my $PASSWORDS, "<", $passwordFile or die "Can't open password file $passwordFile!\n";
	while ( my $entry = <$PASSWORDS> ) {
		chomp $entry;
		next if !$entry or $entry =~ m"^#";
		next if $entry !~ m"(.*)=(.*)";
		my ( $user, $pw ) = ( $1, $2 );
		$user =~ s"(\*?)$"";
		my $admin = $1;
		$user                   = ucfirst lc $user;
		$Users{$user}{password} = $pw;
		$Users{$user}{admin}    = $admin;
	}
	close $PASSWORDS;
}

sub SaveUsers {
	return if !$properties{authentication};
	open my $PASSWORDS, ">", $passwordFile;
	foreach my $user ( sort keys %Users ) {
		print $PASSWORDS "$user$Users{$user}{admin}=$Users{$user}{password}\n";
	}
	close $PASSWORDS;
}

sub ValidateLogin {
	my ( $username, $password ) = @_;
	LoadUsers();
	return FALSE if !exists $Users{$username};
	return FALSE if PasswordHash($password) ne $Users{$username}{password};
	return TRUE;
}

sub ChangePassword {
	my ( $username, $newpassword ) = @_;
	$Users{$username}{password} = PasswordHash($newpassword);
	SaveUsers();
	return TRUE;
}

sub CheckAdminUser {
	my $addAdmin = FALSE;
	if ( !-f $passwordFile ) {    #continue }
		$addAdmin = TRUE;
	} else {
		$addAdmin = TRUE;
		foreach my $user ( keys %Users ) {
			$addAdmin = FALSE if $Users{$user}{admin};
		}
	}
	return if !$addAdmin;
	$Users{'Admin'}{password} = '';
	$Users{'Admin'}{admin}    = '*';
	SaveUsers();
}

# ---------------------------------------------------------------------------------------------
sub CleanupSession {    # cleanup current session
	my ($token) = @_;
	SaveSession($token);
	delete $Tokens{$token};
}

sub SessionCleanup {    # cleanup timedout sessions
	return if !$properties{timeout};
	foreach my $token ( keys %Tokens ) {
		next if $Now{seconds} - $Tokens{$token}{lastaccess} < ( $properties{timeout} * 60 );
		CleanupSession($token);
	}
}

sub ValidateSession {
	my ($token) = @_;
	$CurrentToken = '';
	SessionCleanup();
	return ShowLoginForm('Session timed out, please login.') if !$Tokens{$token};
	$Tokens{$token}{lastaccess} = $Now{seconds};
	$CurrentToken               = $token;
	$Session                    = $sessions{$CurrentToken};
	return TRUE;
}

sub CheckAuthentication {    #object is 'login' or a session ID
	my ( $object ) = @_;
	return NoAuthentication($object) if !$properties{authentication};
	return ShowLoginForm('') if !$object;
	if ( $object eq 'login' ) {
		return HTMLCloseWindow() if $input{cancel};
		my $username    = ucfirst lc $input{username};
		my $password    = $input{password};
		my $newpassword = $input{newpassword};
		return ShowLoginForm('Invalid login!') if !ValidateLogin( $username, $password );
		return HTTPRedirect( GetToken($username) ) if $input{login};
		return ShowLoginForm('Password change failed!') if !ChangePassword( $username, $newpassword );
		return ShowLoginForm("Password changed to '$newpassword'");
	}
	return ValidateSession($object);
}

sub NoAuthentication {
	my ($object) = @_;
	return HTTPRedirect( GetToken($clientIP) ) if !$object;
	$CurrentToken = $object;
	return TRUE;
}

# ---------------------------------------------------------------------------------------------
# End
1;

__END__
