#!/usr/bin/perl
use strict;
use Getopt::Std;
use File::Basename;
use File::Spec;
use FindBin;
use lib "$FindBin::Bin";
use Cwd;
use warnings;

BEGIN {
	my $_ThisScript = $0;
	while ( -l $_ThisScript ) { $_ThisScript = readlink $_ThisScript; }    # follow links
	$_ThisScript = File::Spec->catfile( cwd(), $_ThisScript ) if !File::Spec->file_name_is_absolute($_ThisScript);    # turn it into an absolute path
	chdir( dirname $_ThisScript);                                                                                     # Program directory
}
# general modules
use PDFUtil;
use DateIncrement;
# 'use' hierarchy of Checkbook modules
use Common;
use Multiuser;
use Data;
use HTML;
use Authenticate;
use Configure;
use Queries;
use Reports;
use Pages;
use Process;
use Server;

# ---------------------------------------------------------------------------------------------
# main
SetNow();
LoadConfig();
if ( !LockProgram() ) {
	print STDERR "$Program is already running!\n" if $showStatus;
	StartBrowser();
	exit 0;
}
LoadUsers();
DoBackups();
SetPort();
OpenSocket();
OpenDatabase();
SaveConfig();
StartBrowser();
RunServer();
SaveAllSessions();

# ---------------------------------------------------------------------------------------------
# End
