package Common;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	TRUE FALSE
	$Q
	$DataDir
	$ReportsDir
	%properties
	%defaults
	%Users
	$Program
	$Title
	$Version
	$Shutdown
	$showStatus
	$configFile
	$logFile
	%Now
	%Backups
	$LastBackupTime
	$clientIP
	%input
	@queryResults
	$queryTotal
	$isCurrent
	$isRecurring
	$isLatest
	$isNewAccount
	$isNewStatement
	$isNewTransaction
	$isExistingAccount
	$isExistingStatement
	$isExistingTransaction
	$viewAccounts
	$viewStatements
	$viewTransactions
	$viewQueries
	$viewReports
	%Accounts
	%Statements
	%Transactions
	%Details
	$UserAgent
	$GlobalMessage
	$DownloadFile
	$DownloadContent
	$DownloadContentType
	$DownloadFilename
	LockProgram
	DoBackups
	LoadPropertyFile
	LoadConfig
	SaveConfig
	SetNow
	PageValues
	Log
	PortIsAvailable
);
use strict;
use warnings;
use Date::Calc qw(:all);
use File::Basename;
use File::Path;
use File::Copy;
use Fcntl ':flock';    # import LOCK_* constants

# ---------------------------------------------------------------------------------------------
# constants
use constant TRUE       => 1;
use constant FALSE      => 0;
use constant INPUTDEBUG => 0;

# ---------------------------------------------------------------------------------------------
# public variables
use vars qw(
	$Q
	$DataDir
	$ReportsDir
	%properties
	%defaults
	%Users
	$Program
	$Title
	$Version
	$Shutdown
	$showStatus
	$configFile
	$logFile
	%Now
	%Backups
	$LastBackupTime
	$clientIP
	%input
	@queryResults
	$queryTotal
	$isCurrent
	$isRecurring
	$isLatest
	$isNewAccount
	$isNewStatement
	$isNewTransaction
	$isExistingAccount
	$isExistingStatement
	$isExistingTransaction
	$viewAccounts
	$viewStatements
	$viewTransactions
	$viewQueries
	$viewReports
	%Accounts
	%Statements
	%Transactions
	%Details
	$UserAgent
	$GlobalMessage
	$DownloadFile
	$DownloadContent
	$DownloadContentType
	$DownloadFilename
);
$Q       = '"';
$DataDir = "Data";
mkdir $DataDir, 0777 if !-d $DataDir;
$ReportsDir = "$DataDir/Reports";
mkdir $ReportsDir, 0777 if !-d $ReportsDir;
$Title   = "Checkbook";
$Program = basename $0;
$Program =~ s"\.pl$"";
$configFile           = "$DataDir/$Program.cfg";
$logFile              = "$DataDir/$Program.log";
$Backups{$configFile} = TRUE;
my $lockFile = "$DataDir/$Program.lock";
$GlobalMessage = '';
%defaults      = (
	session => {
		view            => 'transactions',
		transactionsort => 'date',
	},
	properties => {
		authentication => FALSE,
		shutdownonquit => FALSE,
		fieldhelp      => TRUE,
		dynamicport    => FALSE,
		port           => 7777,
		timeout        => 10,            # in minutes
		backupschedule => 4,             # in hours
	},
);
$showStatus = -t STDERR;
$Version    = "";
if ( open my $VERSION, "<", $Program . '.ver' ) {
	$Version = <$VERSION>;
	close $VERSION;
	chomp $Version;
	$Version =~ s" .*"";
}

# ---------------------------------------------------------------------------------------------
my $lockHandle;

sub LockProgram {
	return 0 if !open $lockHandle, ">", $lockFile or !flock( $lockHandle, LOCK_EX | LOCK_NB );
	return 1;
}

END {
	select STDOUT;
	if ($lockHandle) {
		flock( $lockHandle, LOCK_UN );
		close $lockHandle;
	}
	unlink $lockFile if -e $lockFile;
}

# ---------------------------------------------------------------------------------------------
sub DoBackups {
	my $backuproot = "${DataDir}/Backup";
	mkdir $backuproot, 0777 if !-d $backuproot;
	my $BackupDir = sprintf "%s/%.4d%.2d%.2d-%.2d%.2d%.2d", $backuproot, $Now{year}, $Now{month}, $Now{day}, $Now{hour}, $Now{minute}, $Now{second};
	return if -d $BackupDir;
	mkdir $BackupDir, 0777;
	foreach my $file ( sort keys %Backups ) {
		copy( $file, $BackupDir ) if -f $file;
	}
	$LastBackupTime = $Now{seconds};
	CleanupOldBackups($backuproot);
}

sub CleanupOldBackups {    # keep last backup of each month, plus all backups from this and last month
	my ($backuproot) = @_;
	my %OldBackups;
	my %LastOfMonth;
	my $ThisYearMonth = sprintf "%.4d%.2d", $Now{year}, $Now{month};
	my $LastMonth     = $Now{month} - 1;
	my $LastYear      = $Now{year};
	if ( !$LastMonth ) {
		$LastMonth = 12;
		$LastYear--;
	}
	my $LastYearMonth = sprintf "%.4d%.2d", $LastYear, $LastMonth;
	foreach my $backupdir ( glob("$backuproot/*") ) {
		next if !-d $backupdir;
		my $backup = basename $backupdir;
		next if $backup !~ m"(\d\d\d\d)(\d\d)\d\d-\d\d\d\d\d\d";
		my ( $backupYear, $backupMonth ) = ( $1, $2 );
		my $backupYearMonth = $backupYear . $backupMonth;
		next if $backupYearMonth eq $ThisYearMonth or $backupYearMonth eq $LastYearMonth;
		$OldBackups{$backup}           = $backupYearMonth;
		$LastOfMonth{$backupYearMonth} = $backup if !$LastOfMonth{$backupYearMonth};
		$LastOfMonth{$backupYearMonth} = $backup if $backup gt $LastOfMonth{$backupYearMonth};
	}
	foreach my $backup ( sort keys %OldBackups ) {
		my $yearMonth   = $OldBackups{$backup};
		my $lastOfMonth = $LastOfMonth{$yearMonth};
		next if $backup eq $lastOfMonth;
		rmtree( "$backuproot/$backup", FALSE, FALSE );
	}
}

# ---------------------------------------------------------------------------------------------
sub LoadPropertyFile {
	my ( $file, $hashref ) = @_;
	return FALSE if !-f $file;
	open my $FILE, "<", $file;
	while ( my $entry = <$FILE> ) {
		chomp $entry;
		next if $entry !~ m"(.*)=(.*)";
		$$hashref{$1} = $2;
	}
	close $FILE;
}

# ---------------------------------------------------------------------------------------------
sub LoadConfig {
	foreach my $property ( keys %{ $defaults{properties} } ) {
		$properties{$property} = $defaults{properties}{$property};
	}
	LoadPropertyFile( $configFile, \%properties );
}

sub SaveConfig {
	my ($fromInput) = @_;
	open my $CFG, ">", $configFile;
	foreach my $property ( sort keys %properties ) {
		next if $properties{$property} eq 'new';
		$properties{$property} = $input{$property} if $fromInput and exists $input{$property};
		print $CFG "$property=$properties{$property}\n";
	}
	close $CFG;
}

# ---------------------------------------------------------------------------------------------
sub SetNow {
	$Now{datetime} = [ localtime() ];
	( $Now{second}, $Now{minute}, $Now{hour}, $Now{day}, $Now{month}, $Now{year}, $Now{weekday} ) = @{ $Now{datetime} };
	$Now{month}++;
	$Now{year} += 1900;
	$Now{WeekDay} = ( 'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat' )[ $Now{weekday} ];
	$Now{date} = sprintf "%.4d-%.2d-%.2d", $Now{year}, $Now{month}, $Now{day};
	$Now{seconds} = time();
	( $Now{yesteryear}, $Now{yestermonth}, $Now{yesterday} ) = Add_Delta_YMD( $Now{year}, $Now{month}, $Now{day}, 0, 0, -1 );
}

# ---------------------------------------------------------------------------------------------
sub PageValues {
	my ($content) = @_;
	%input = ();
	return if !$content;
	my @content = ( split( m"[\n&]", $content ) );
	foreach my $entry (@content) {
		$entry =~ tr/+/ /;
		$entry =~ s/%([a-fA-F0-9]{2,2})/chr(hex($1))/eg;
		$entry =~ s/<!--(.|\n)*-->//g;
		next if $entry !~ m"^([^=]*)=(.*)";
		my ( $property, $value ) = ( $1, $2 );
		$input{$property} = $value;
		print STDERR "input $property=$value\n" if INPUTDEBUG;
	}
}

# ---------------------------------------------------------------------------------------------
sub Log {
	return;    # logging disabled
	my (@logs) = @_;
	my $log = join( " ", @logs ) . "\n";
	open my $LOG, ">>", $logFile;
	print $LOG $log;
	close $LOG;
}

# ---------------------------------------------------------------------------------------------
sub PortIsAvailable {
	my ($port) = @_;
	my @netstat = `netstat -a -n`;
	my $portInUse = grep m/[:.]$port\s/, grep !/TIME_WAIT/, @netstat;
	return !$portInUse;
}

# ---------------------------------------------------------------------------------------------
# End
1;
