package Configure;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	Configure
	Users
);
use strict;
use warnings;
use Sys::Hostname;
use Common;
use Multiuser;
use HTML;
use Authenticate;

# ---------------------------------------------------------------------------------------------
# variables
use vars qw ($AuthenticationActivated);
$AuthenticationActivated = 0;

# ---------------------------------------------------------------------------------------------
sub ShowConfigureForm {
	my ($message) = @_;
	HTTPHeader();
	HTMLHeader();
	print HTMLStart("body style='background-color:lightgray;' ");
	print HTMLStart('center');
	print HTMLStartEnd( "h1 style='background-color:pink'",       "$Title <small><small>Version $Version</small></small><hr class='nomargin'>" );
	print HTMLStartEnd( "h2 style='background-color:lightblue;'", "Program Configuration" );
	print HTMLStart("div");
	print HTMLStart('table border');
	print HTMLStart("form method=POST action='configure' ");
	print HTMLStartEnd( "tr style='background-color:red;'", HTMLEnclosed( "th colspan=2", "Don't change anything unless you know what you're doing!" ) );
	my $authchecked = $properties{authentication} ? 'checked' : '';
	my $quitchecked = $properties{shutdownonquit} ? 'checked' : '';
	my $portchecked = $properties{dynamicport}    ? 'checked' : '';
	my $helpchecked = $properties{fieldhelp}      ? 'checked' : '';
	print HTMLStartEnd( "tr", HTMLEnclosed( "th style='background-color:lightblue;'", 'Authentication?' ),                HTMLEnclosed( "td", HTMLEnclosed("input type=checkbox name=authentication value=1 $authchecked") ) );
	print HTMLStartEnd( "tr", HTMLEnclosed( "th style='background-color:lightblue;'", 'Quit Shuts Down? (for testing)' ), HTMLEnclosed( "td", HTMLEnclosed("input type=checkbox name=shutdownonquit value=1 $quitchecked") ) );
	print HTMLStartEnd( "tr", HTMLEnclosed( "th style='background-color:lightblue;'", 'Show Field Help (Tool Tips)?' ),   HTMLEnclosed( "td", HTMLEnclosed("input type=checkbox name=fieldhelp value=1 $helpchecked") ) );
	print HTMLStartEnd( 'tr', HTMLEnclosed( "th style='background-color:lightblue;'", 'Backup Schedule (hours)' ),        HTMLEnclosed( 'td', HTMLEnclosed("input type=text name=backupschedule value=$properties{backupschedule}") ) );
	print HTMLStartEnd( 'tr', HTMLEnclosed( "th style='background-color:lightblue;'", 'Session Timeout (minutes)' ),      HTMLEnclosed( 'td', HTMLEnclosed("input type=text name=timeout value=$properties{timeout}") ) );
	my $BrowserSelect = "";
	my %Browsers = ( default => 'System Default', iexplore => 'Internet Explorer', firefox => 'Firefox', chrome => 'Chrome' );
	foreach my $browser ( 'default', 'iexplore', 'firefox', 'chrome' ) {
		next if $browser ne 'default' and !$properties{$browser};
		my $selected = $properties{browser} eq $browser ? 'selected' : '';
		$BrowserSelect .= HTMLEnclosed( "option value='$browser' $selected", $Browsers{$browser} );
	}
	print HTMLStartEnd( 'tr', HTMLEnclosed( "th style='background-color:lightblue;'", 'Browser' ), HTMLEnclosed( 'td', HTMLEnclosed( 'select name=browser', $BrowserSelect ) ) );
	print HTMLStartEnd( 'tr', HTMLEnclosed( "th style='background-color:lightblue;'", 'Port Number (program will shutdown if changed)' ),         HTMLEnclosed( 'td', HTMLEnclosed("input type=text name=port value=$properties{port}") ) );
	print HTMLStartEnd( "tr", HTMLEnclosed( "th style='background-color:lightblue;'", 'Dynamic Port value?  (enable to handle port conflicts)' ), HTMLEnclosed( "td", HTMLEnclosed("input type=checkbox name=dynamicport value=1 $portchecked") ) );
	print HTMLStartEnd( 'tr', HTMLEnclosed( 'th colspan=2',                           HTMLEnclosed("input type=submit name='save' value='Save'") ) );
	print HTMLStartEnd( 'tr', HTMLEnclosed( 'th colspan=2',                           HTMLEnclosed("input type=submit name='cancel' value='Cancel'") ) );
	print HTMLStartEnd( 'tr', HTMLEnclosed( 'td colspan=2', HTMLEnclosed("input type=submit name='shutdown' value='Shutdown $Title'") ) );
	print HTMLStartEnd( "tr style='background-color:red;'", HTMLEnclosed( "th colspan=2", $message ) ) if $message;
	print HTMLEnd('form');
	print HTMLEnd('table');
	print HTMLEnd('div');
	print HTMLEnd('center');
	print HTMLEnd('body');
	print HTMLEnd('html');
	return FALSE;
}

# ---------------------------------------------------------------------------------------------
sub ValidateConfig {
	return "Backup Schedule must be a number of hours."     if $input{backupschedule} !~ m"^\d+$";
	return "Session Timeout must be a number of minutes."   if $input{timeout} !~ m"^\d+$";
	return "Port Number must be a number less then 65535."  if $input{port} !~ m"^\d+$" or $input{port} > 65535;
	return "Port $input{port} is not available."            if $input{port} ne $properties{port} and !CheckPort( $input{port} );
	foreach my $property ( 'authentication', 'shutdownonquit', 'fieldhelp', 'dynamicport' ) {
		$input{$property} = !!$input{$property};                                                 # force zero or 1
	}
	$input{'url'} = 'http://' . hostname() . ':' . $input{port};
	$GlobalMessage = 'Program Configuration updated.';
	$AuthenticationActivated = 1 if $input{authentication} and !$properties{authentication};
	return '';
}

# ---------------------------------------------------------------------------------------------
sub Configure {
	if ( $input{shutdown} ) {
		$Shutdown = TRUE;
		return HTMLCloseWindow();
	}
	return HTTPRedirect('') if $input{cancel};
	if ( $input{save} ) {
		my $Result = ValidateConfig();
		return ShowConfigureForm($Result) if $Result;
		$Shutdown = TRUE if $input{port} ne $properties{port};
		SaveConfig(TRUE);
		CheckAdminUser() if $properties{authentication};
		return HTMLCloseWindow() if $Shutdown;
		if ($AuthenticationActivated) {
			$AuthenticationActivated = 0;
			LoadUsers();
			return HTTPRedirect('users');
		}
		return HTTPRedirect('');
	}
	return ShowConfigureForm();
}

# ---------------------------------------------------------------------------------------------
sub ShowUsersForm {
	my ($message) = @_;
	HTTPHeader();
	HTMLHeader();
	print HTMLStart("body style='background-color:lightgray;' ");
	print HTMLStart('center');
	print HTMLStartEnd( "h1 style='background-color:pink'", "$Title <small><small>Version $Version</small></small><hr class='nomargin'>" );
	print HTMLStartEnd( "h2",                               "User Maintenance" );
	print HTMLStart("div");
	print HTMLStart('table border');
	print HTMLStart("form method=POST action='users' ");
	print HTMLStartEnd( 'tr', HTMLEnclosed( 'th', 'User' ), HTMLEnclosed( 'th colspan=3', 'Action' ) );

	foreach my $user ( sort keys %Users ) {
		print HTMLStart('tr');
		print HTMLStartEnd( "th style='background-color:lightblue;'", $user )                                                                                       if $Users{$user}{admin};
		print HTMLStartEnd( 'th',                                     $user )                                                                                       if !$Users{$user}{admin};
		print HTMLStartEnd( 'td',                                     HTMLEnclosed("input type=submit name='user_delete_$user' value='Delete User'") );
		print HTMLStartEnd( 'td',                                     HTMLEnclosed("input type=submit name='user_resetpassword_$user' value='Reset Password'") );
		print HTMLStartEnd( 'td',                                     HTMLEnclosed("input type=submit name='user_grantadmin_$user' value='Grant Admin Access'") )   if !$Users{$user}{admin};
		print HTMLStartEnd( 'td',                                     HTMLEnclosed("input type=submit name='user_revokeadmin_$user' value='Revoke Admin Access'") ) if $Users{$user}{admin};
		print HTMLEnd('tr');
	}
	print HTMLStart('tr');
	print HTMLEnd('form');
	print HTMLEnd('table');
	print HTMLEnd('div');
	print HTMLStart("div");
	print HTMLStart("form method=POST action='users' ");
	print HTMLSingleIndented('br'), HTMLSingleIndented('hr'), HTMLSingleIndented('br');
	print HTMLStartEnd("input type=text name=user");
	print HTMLStartEnd("input type=submit name='newuser' value='Add User'");
	print HTMLSingleIndented('br'), HTMLSingleIndented('hr'), HTMLSingleIndented('br');
	print HTMLStartEnd("input type=submit name='quit' value='Quit'");
	print HTMLSingleIndented('br');
	print HTMLStartEnd("input type=submit name='shutdown' value='Shutdown $Title'");
	print HTMLSingleIndented('br'), HTMLSingleIndented('hr'), HTMLSingleIndented('br'), HTMLStartEnd( "div style='background-color:red;'", $message ) if $message;
	print HTMLEnd('form');
	print HTMLEnd('div');
	print HTMLEnd('center');
	print HTMLEnd('body');
	print HTMLEnd('html');
	return FALSE;
}

# ---------------------------------------------------------------------------------------------
sub Users {
	if ( $input{shutdown} ) {
		$Shutdown = TRUE;
		return HTMLCloseWindow();
	}
	return HTTPRedirect('') if $input{quit};
	if ( $input{newuser} ) {
		return ShowUsersForm("Please enter a Username to add.") if !$input{user};
		my $user = ucfirst lc $input{user};
		return ShowUsersForm("User '$user' already exists.") if exists $Users{$user};
		$Users{$user}{password} = '';
		$Users{$user}{admin}    = '';
		SaveUsers();
		return ShowUsersForm("User '$user' added.");
	}
	foreach my $input ( keys %input ) {
		next if $input !~ m"^user_(.*)_(.*)$";
		my ( $action, $user ) = ( $1, $2 );
		if ( $action eq 'delete' ) {
			delete $Users{$user};
			SaveUsers();
			return ShowUsersForm("User '$user' removed.");
		}
		if ( $action eq 'resetpassword' ) {
			$Users{$user}{password} = '';
			SaveUsers();
			return ShowUsersForm("User '$user' password is now blank.");
		}
		if ( $action eq 'grantadmin' ) {
			$Users{$user}{admin} = '*';
			SaveUsers();
			return ShowUsersForm("User '$user' now has admin access.");
		}
		if ( $action eq 'revokeadmin' ) {
			$Users{$user}{admin} = '';
			SaveUsers();
			return ShowUsersForm("User '$user' no longer has admin access.");
		}
	}
	return ShowUsersForm();
}

# ---------------------------------------------------------------------------------------------
# End
1;
