package Data;
use constant DEBUG => 0;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	REVERSE RECALCULATE
	$database
	$transactionDebug
	%TransactionTypes
	%StatementTotalTypes
	%AccountTypes
	$defaultAccountType
	$defaultTransactionType
	%DefaultTransactionTypes
	@accountFields
	@statementFields
	@transactionFields
	@detailFields
	%numericFields
	%moneyFields
	%columnHeaders
	%columnFormats
	%emptyFields
	%Fields
	%Records
	%FieldHelp
	@Tables
	$addAccount
	$addStatement
	$addTransaction
	$addDetail
	$fixAvailable
	@deletedTransactionFields
	OpenDatabase
	StartDBTransaction
	FinishDBTransaction
	CloseDatabase
	AddAccount
	AddAccountEdit
	AddStatement
	AddStatementNew
	AddTransaction
	AddTransactionEdit
	AddTransactionRecord
	AddDetail
	AddDetailEdit
	Amount
	ValidDate
	ValidAmount
	ValidNumber
	FormatAmount
	FormatDate
	FormatCheck
	FormatDesc
	ManualStatementBalance
	UpdateBalance
	ApplyTransaction
	RecalculateStatementBalances
	RecalculateStatement
	RecalculateBalances
	EditRecord
	NewRecord
	InitialRecordFields
	SortItem
	SortItem2
);
use strict;
use warnings;
use DBI;
use Tie::DBI;
use Date::Calc qw(:all);
use Time::Local;
use Common;
use Multiuser;

# ---------------------------------------------------------------------------------------------
# constants
use constant REVERSE     => -1;
use constant RECALCULATE => 1;

# ---------------------------------------------------------------------------------------------
# variables
use vars qw(
	$database
	$transactionDebug
	%TransactionTypes
	%StatementTotalTypes
	%AccountTypes
	$defaultAccountType
	$defaultTransactionType
	%DefaultTransactionTypes
	@accountFields
	@statementFields
	@transactionFields
	@detailFields
	%numericFields
	%moneyFields
	%columnHeaders
	%columnFormats
	%emptyFields
	%Fields
	%Records
	%FieldHelp
	@Tables
	$addAccount
	$addStatement
	$addTransaction
	$addDetail
	@deletedTransactionFields
);
$transactionDebug = FALSE;

# c,d are supposed to be credit,debit, but ended up more like charge,deposit.  d=additions, c=subtractions
%TransactionTypes = (    # ^a=addition, s=subtraction
	'd1' => 'Deposit',
	'd2' => 'Payment',
	'd3' => 'Credit',
	'd4' => 'Interest',
	'd5' => 'Transfer In',
	'c1' => 'Charge',
	'c2' => 'ATM',
	'c3' => 'Check',
	'c4' => 'Transfer Out',
	'c5' => 'Withdrawal',
	'c6' => 'Interest Charges',
);
%StatementTotalTypes = (
	'd' => [ 'beginningbalance=Beginning Balance', 'additions1=Deposits/Credits', 'subtractions1=Withdrawals/Debits', 'endingbalance=Ending Balance', ],
	'c' => [ 'beginningbalance=Previous Balance',  'subtractions1=Payments',      'subtractions2=Other Credits',      'additions1=Purchases/Debits', 'additions2=Interest Charges', 'endingbalance=New Balance', ],
);
%AccountTypes = (    # ^d=Dr account balance, c=Cr account balance
	'd' => {
		label            => 'Checking/Savings',
		transactiontypes => {
			'd1' => 'transactionadditions1',
			'd3' => 'transactionadditions1',
			'd4' => 'transactionadditions1',
			'd5' => 'transactionadditions1',
			'c1' => 'transactionsubtractions1',
			'c2' => 'transactionsubtractions1',
			'c3' => 'transactionsubtractions1',
			'c4' => 'transactionsubtractions1',
			'c5' => 'transactionsubtractions1',
		},
	},
	'c' => {
		label            => 'Credit Card',
		transactiontypes => {
			'd2' => 'transactionsubtractions1',
			'd3' => 'transactionsubtractions2',
			'c1' => 'transactionadditions1',
			'c6' => 'transactionadditions2',
		},
	},
);
$defaultAccountType      = 'c';
$defaultTransactionType  = 'c1';
%DefaultTransactionTypes = ( 'd' => 'd3', 'c' => 'c1' );
%numericFields           = (
	'account' => {
		'archived'     => 1,
		'lastcheck'    => 1,
		'interestrate' => 1,
	},
	'statement'   => { 'statementdate' => 1, },
	'transaction' => {
		'cleared'     => 1,
		'checknumber' => 1,
		'date'        => 1,
	},
	'detail' => {},
);
%moneyFields = (
	'account' => {
		'balance'        => 1,
		'clearedbalance' => 1,
		'creditlimit'    => 1,
	},
	'statement' => {
		'beginningbalance'         => 1,
		'statementadditions1'      => 1,
		'statementadditions2'      => 1,
		'statementsubtractions1'   => 1,
		'statementsubtractions2'   => 1,
		'statementendingbalance'   => 1,
		'transactionadditions1'    => 1,
		'transactionadditions2'    => 1,
		'transactionsubtractions1' => 1,
		'transactionsubtractions2' => 1,
		'transactionendingbalance' => 1,
		'discrepancyadditions1'    => 1,
		'discrepancyadditions2'    => 1,
		'discrepancysubtractions1' => 1,
		'discrepancysubtractions2' => 1,
		'discrepancyendingbalance' => 1,
	},
	'transaction' => { 'amount' => 1, },
	'detail'      => { 'amount' => 1, },
);
%columnHeaders = (
	'date'          => sprintf( "%-10s", 'Date' ) . '  ',
	'checknumber'   => sprintf( "%6s",   'Check#' ) . '  ',
	'amount'        => sprintf( "%8s",   'Amount' ) . '  ',
	'description'   => sprintf( "%-30s", 'Description' ) . '  ',
	'statementdate' => sprintf( "%-10s", 'Cleared' ) . '  ',
	'debits'        => sprintf( "%9s",   'Debits' ) . '  ',
	'credits'       => sprintf( "%9s",   'Credits' ) . '  ',
	'balance'       => sprintf( "%9s",   'Balance' ) . '  ',
);
%columnFormats = (
	'date'   => "%-10s  ",
	'check'  => "%6d  ",
	'amount' => "%9.2f  ",
	'desc'   => "%-30s  ",
);
%emptyFields = (
	'date'   => ' ' x 12,
	'check'  => ' ' x 8,
	'amount' => ' ' x 11,
	'desc'   => ' ' x 32,
	'balance' => ' ' x 11,
);
%FieldHelp = (
	account => {
		accountname    => 'Enter a name for this account.',
		accounttype    => 'Select a type for this account.',
		accountnumber  => 'Enter the account number.',
		routingnumber  => 'Enter the bank routing number, if any.',
		cardnumber     => 'If this is a checking account with a debit card,<br>enter the debit card number.',
		expirationdate => 'If this account has a debit or credit card, enter the expiration date.',
		securitycode   => '(Optional) If this account has a debit or credit card, enter the 3-digit CVV.',
		creditlimit    => 'Enter the credit limit for the account.',
		interestrate   => 'Enter the interest rate for the account.',
	},
	statement => {
		statementdate          => 'Enter the statement date for this statement.<br>The date can be left blank until the statement is finalized.',
		beginningbalance       => 'Enter the beginning balance for the statement.',
		statementadditions1    => 'Enter total deposits or payments for this statement.',
		statementadditions2    => 'Enter total other credits for this statement.',
		statementsubtractions1 => 'Enter total charges for this statement.',
		statementsubtractions2 => 'Enter interest charges for this statement.',
	},
	transaction => {
		transactiontype => 'Select a transaction type.',
		amount          => 'Enter the amount of the transaction.',
		checknumber     => 'If this is a check, enter the check number.',
		date            => 'Enter a date in one of these formats:<br>  YYYY-MM-DD, MM/DD/YY, MM/DD, DD.<br>' . 'Use blank or zero for "today"; use a minus sign for "yesterday".',
		description     => 'Enter the description or payee for this transaction.',
		description2    => 'Enter additional description for this transaction.',
		description3    => 'Enter additional description for this transaction.',
		increment       => 'Enter an increment for this recurring transaction; choices are: annual, monthly, semimonthly, weekly, daily.  ' . 'The default is monthly.  You can also include a number.  For example, weekly4 would be every 4 weeks.',
	},
	detail => {
		description => 'Enter a description for this entry.',
		amount      => 'Enter an amount for this entry.',
	},
);
@deletedTransactionFields = ( 'transactiontype', 'amount', 'checknumber', 'date', 'description', 'description2', 'description3' );

# ---------------------------------------------------------------------------------------------
# Open/Close
use vars qw ($dbopen);
my $DataFile = "$DataDir/$Program.db";
$Backups{$DataFile} = TRUE;

sub OpenDatabase {
	my ($AutoCommit) = @_;
	CreateDatabase($DataFile) if !-f $DataFile;
	$AutoCommit = 1 if !defined $AutoCommit;
	DoBackups();
	$database = DBI->connect( "dbi:SQLite:dbname=$DataFile", "", "" );
	$dbopen = TRUE;
	$database->do("PRAGMA foreign_keys = ON");
	$database->{AutoCommit} = $AutoCommit;
	DatabaseUpgrade();
	tie %Accounts,     'Tie::DBI', { db => $database, table => 'accounts',     key => 'key', CLOBBER => 2, AUTOCOMMIT => $AutoCommit, };
	tie %Statements,   'Tie::DBI', { db => $database, table => 'statements',   key => 'key', CLOBBER => 2, AUTOCOMMIT => $AutoCommit, };
	tie %Transactions, 'Tie::DBI', { db => $database, table => 'transactions', key => 'key', CLOBBER => 2, AUTOCOMMIT => $AutoCommit, };
	tie %Details,      'Tie::DBI', { db => $database, table => 'details',      key => 'key', CLOBBER => 2, AUTOCOMMIT => $AutoCommit, };
	@accountFields = grep ~m"^key$", tied(%Accounts)->fields;
	$Fields{account}  = \@accountFields;
	$Records{account} = \%Accounts;
	push @Tables, ('account');
	@statementFields = grep ~m"^key$", tied(%Statements)->fields;
	$Fields{statement}  = \@statementFields;
	$Records{statement} = \%Statements;
	push @Tables, ('statement');
	@transactionFields = grep ~m"^key$", tied(%Transactions)->fields;
	$Fields{transaction}  = \@transactionFields;
	$Records{transaction} = \%Transactions;
	push @Tables, ('transaction');
	@detailFields = grep ~m"^key$", tied(%Details)->fields;
	$Fields{detail}  = \@detailFields;
	$Records{detail} = \%Details;
	push @Tables, ('detail');
	$addAccount = $database->prepare( "
	INSERT INTO
	accounts (accountname, accounttype, accountnumber, routingnumber, cardnumber, expirationdate, securitycode, creditlimit, interestrate, archived, balance, clearedbalance, lastcheck, currentstatement, recurringstatement, lateststatement) 
	VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
	" );
	$addStatement = $database->prepare( "
	INSERT INTO
	statements (account, statementdate, beginningbalance, statementadditions1, statementadditions2, statementsubtractions1, statementsubtractions2, statementendingbalance, transactionadditions1, transactionadditions2, transactionsubtractions1, transactionsubtractions2, transactionendingbalance, discrepancyadditions1, discrepancyadditions2, discrepancysubtractions1, discrepancysubtractions2, discrepancyendingbalance )
	VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
	" );
	$addTransaction = $database->prepare( "
	INSERT INTO
	transactions (statementkey, cleared, transactiontype, amount, checknumber, date, description, description2, description3, increment)
	VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
	" );
	$addDetail = $database->prepare( "
	INSERT INTO
	details (transactionkey, description, amount)
	VALUES (?, ?, ?)
	" );
}

END {
	$database->disconnect if $dbopen;
}

sub CreateDatabase {
	my ($datafile) = @_;
	return if -f $datafile;
	my $dbh = DBI->connect( "dbi:SQLite:dbname=$datafile", "", "" );
	$dbh->do("PRAGMA foreign_keys = ON");
	$dbh->{AutoCommit} = 1;
	my $CreateAccount = "
CREATE TABLE accounts (
	key						INTEGER PRIMARY KEY,
	accountname				TEXT UNIQUE,
	accounttype				TEXT NOT NULL,
	accountnumber			TEXT,
	routingnumber           TEXT,
	cardnumber				TEXT,
	expirationdate			TEXT,
	securitycode			TEXT,
	creditlimit				SMALLMONEY NOT NULL,
	interestrate			SMALLMONEY NOT NULL,
	archived				BIT,
	balance					SMALLMONEY NOT NULL,
	clearedbalance			SMALLMONEY NOT NULL,
	lastcheck				INTEGER,
	currentstatement		INTEGER,
	recurringstatement		INTEGER,
	lateststatement			INTEGER
)";
	my $CreateStatement = "
CREATE TABLE statements (
	key							INTEGER PRIMARY KEY,
	account						INTEGER NOT NULL REFERENCES accounts(key) ON DELETE CASCADE,
	statementdate				DATE NOT NULL,
	beginningbalance			SMALLMONEY,
	statementadditions1			SMALLMONEY,
	statementadditions2			SMALLMONEY,
	statementsubtractions1		SMALLMONEY,
	statementsubtractions2		SMALLMONEY,
	statementendingbalance		SMALLMONEY,
	transactionadditions1		SMALLMONEY,
	transactionadditions2		SMALLMONEY,
	transactionsubtractions1	SMALLMONEY,
	transactionsubtractions2	SMALLMONEY,
	transactionendingbalance	SMALLMONEY,
	discrepancyadditions1		SMALLMONEY,
	discrepancyadditions2		SMALLMONEY,
	discrepancysubtractions1	SMALLMONEY,
	discrepancysubtractions2	SMALLMONEY,
	discrepancyendingbalance	SMALLMONEY
)";
	my $CreateTransaction = "
CREATE TABLE transactions (
	key					INTEGER PRIMARY KEY,
	statementkey		INTEGER NOT NULL REFERENCES statements(key) ON DELETE CASCADE,
	cleared				BIT,
	transactiontype		TEXT,
	amount				SMALLMONEY NOT NULL,
	checknumber			INTEGER,
	date				DATE NOT NULL,
	description			TEXT NOT NULL,
	description2		TEXT,
	description3		TEXT,
	increment           TEXT
)";
	my $CreateDetail = "
CREATE TABLE details (
	key					INTEGER PRIMARY KEY,
	transactionkey		INTEGER NOT NULL REFERENCES transactions(key) ON DELETE CASCADE,
	description			TEXT NOT NULL,
	amount				SMALLMONEY NOT NULL
)";

	foreach my $Create ( $CreateAccount, $CreateStatement, $CreateTransaction, $CreateDetail ) {
		my ($Table) = $Create =~ m"TABLE (.*)\(";
		print STDERR "\ncreating table $Table\n" if $showStatus;
		my $rows = $dbh->do($Create);
		if ( $dbh->err ) {
			print STDERR "Error ", $dbh->err, ": ", $dbh->errstr, "\n" if $dbh->err and $showStatus;
			next;
		}
		print STDERR "table $Table created.\n" if $showStatus;
	}
	$dbh->disconnect;
}

# ---------------------------------------------------------------------------------------------
sub DatabaseUpgrade {
	my $columnStatement = $database->column_info( undef, undef, 'accounts', undef );
	my $columns = $columnStatement->fetchall_hashref('COLUMN_NAME');
	if ( !$$columns{routingnumber} ) {
		StartDBTransaction();
		$database->do("ALTER TABLE accounts ADD COLUMN routingnumber TEXT");
		print STDERR "Error ", $database->err, ": ", $database->errstr, "\n" if $database->err and $showStatus;
		FinishDBTransaction();
	}
	if ( !$$columns{creditlimit} ) {
		StartDBTransaction();
		$database->do("ALTER TABLE accounts ADD COLUMN creditlimit TEXT");
		$database->do("ALTER TABLE accounts ADD COLUMN interestrate TEXT");
		print STDERR "Error ", $database->err, ": ", $database->errstr, "\n" if $database->err and $showStatus;
		FinishDBTransaction();
	}
}

# ---------------------------------------------------------------------------------------------
sub StartDBTransaction {
	$database->{AutoCommit} = 0;
}

sub FinishDBTransaction {
	$database->commit;
	$database->{AutoCommit} = 1;
}

# ---------------------------------------------------------------------------------------------
sub CloseDatabase {
	$database->disconnect;
	$dbopen = FALSE;
}

# ---------------------------------------------------------------------------------------------
sub AddAccount {
	my ( $accountname, $accounttype, $accountnumber, $routingnumber, $cardnumber, $expirationdate, $securitycode, $creditlimit, $interestrate ) = @_;
	my $rows = $addAccount->execute( $accountname, $accounttype, $accountnumber, $routingnumber, $cardnumber, $expirationdate, $securitycode, $creditlimit, $interestrate, 0, 0, 0, 0, 0, 0, 0 );
	my $error = $addAccount->errstr;
	if ($error) {
		$Session->{current}{error} = "Error adding account record:\n$error";
		return '';
	}
	return $database->last_insert_id( undef, undef, 'accounts', 'key' );
}

sub AddAccountEdit {
	return AddAccount(
		$Session->{edit}{account}{accountname},    $Session->{edit}{account}{accounttype},  $Session->{edit}{account}{accountnumber}, $Session->{edit}{account}{routingnumber}, $Session->{edit}{account}{cardnumber},
		$Session->{edit}{account}{expirationdate}, $Session->{edit}{account}{securitycode}, $Session->{edit}{account}{creditlimit},   $Session->{edit}{account}{interestrate}
	);
}

# ---------------------------------------------------------------------------------------------
sub AddStatement {
	my ( $account, $statementdate, $beginningbalance, $statementadditions1, $statementadditions2, $statementsubtractions1, $statementsubtractions2 ) = @_;
	my $statementendingbalance = $beginningbalance + $statementadditions1 + $statementadditions2 - $statementsubtractions1 - $statementsubtractions2;
	my $rows                   = $addStatement->execute( $account, $statementdate, $beginningbalance, $statementadditions1, $statementadditions2, $statementsubtractions1, $statementsubtractions2, $statementendingbalance, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
	my $error                  = $addStatement->errstr;
	if ($error) {
		$Session->{current}{error} = "Error adding statement record:\n$error";
		return '';
	}
	return $database->last_insert_id( undef, undef, 'statements', 'key' );
}

sub AddStatementNew {
	my ( $account, $balance ) = @_;
	return AddStatement( $account, 0, $balance, 0, 0, 0, 0 );
}

# ---------------------------------------------------------------------------------------------
sub AddTransaction {
	my ( $statement, $cleared, $transactiontype, $amount, $checknumber, $date, $description, $description2, $description3, $increment ) = @_;
	my $rows = $addTransaction->execute( $statement, $cleared, $transactiontype, $amount, $checknumber, $date, $description, $description2, $description3, $increment, );
	my $error = $addTransaction->errstr;
	if ($error) {
		$Session->{current}{error} = "Error adding transaction record:\n$error";
		return '';
	}
	return $database->last_insert_id( undef, undef, 'transactions', 'key' );
}

sub AddTransactionEdit {
	my ($statement) = @_;
	return AddTransaction(
		$statement,                          0,                                          $Session->{edit}{transaction}{transactiontype}, $Session->{edit}{transaction}{amount},       $Session->{edit}{transaction}{checknumber},
		$Session->{edit}{transaction}{date}, $Session->{edit}{transaction}{description}, $Session->{edit}{transaction}{description2},    $Session->{edit}{transaction}{description3}, $Session->{edit}{transaction}{increment}
	);
}

sub AddTransactionRecord {
	my ( $statement, $transaction ) = @_;
	return AddTransaction(
		$statement,                                0,                                 $Transactions{$transaction}{transactiontype}, $Transactions{$transaction}{amount},
		$Transactions{$transaction}{checknumber},  $Transactions{$transaction}{date}, $Transactions{$transaction}{description},     $Transactions{$transaction}{description2},
		$Transactions{$transaction}{description3}, ''
	);
}

# ---------------------------------------------------------------------------------------------
sub AddDetail {
	my ( $transaction, $description, $amount ) = @_;
	my $rows = $addDetail->execute( $transaction, $description, $amount );
	my $error = $addDetail->errstr;
	if ($error) {
		$Session->{current}{error} = "Error adding detail record:\n$error";
		return '';
	}
	return $database->last_insert_id( undef, undef, 'details', 'key' );
}

sub AddDetailEdit {
	my ($transaction) = @_;
	return AddDetail( $transaction, $Session->{edit}{detail}{description}, $Session->{edit}{detail}{amount} );
}

# ---------------------------------------------------------------------------------------------
sub Amount {
	my ($amount) = @_;
	return "0" if !$amount;
	$amount =~ s"[, \$]""g;
	return $amount if $amount !~ m"^[0-9.-]*$";
	return "0" if abs($amount) < 0.009;
	$amount = sprintf "%.2f", $amount;
	return $amount;
}

# ---------------------------------------------------------------------------------------------
sub ReformatDate {
	my ($dateRef) = @_;
	if ( !$$dateRef ) {    # 0 or empty = today
		$$dateRef = sprintf "%.4d-%.2d-%.2d", $Now{year}, $Now{month}, $Now{day};
		return 1;
	}
	$dateRef =~ s" ""g;
	return 1 if $$dateRef =~ m"^\d\d\d\d-\d\d-\d\d$";                  # yyyy-mm-dd
	return 1 if $$dateRef =~ s"^(\d\d\d\d)(\d\d)(\d\d)$"$1-$2-$3";
	return 1 if $$dateRef =~ s"^(\d\d\d\d)/(\d\d)/(\d\d)$"$1-$2-$3";
	if ( $$dateRef =~ m"^(\d?\d)[/-](\d?\d)[/-](\d\d)$" ) {            # mm/dd/yy
		$$dateRef = sprintf "20%.2d-%.2d-%.2d", $3, $1, $2;
		return 1;
	}
	if ( $$dateRef =~ m"^(\d?\d)[/-](\d?\d)[/-](\d?\d?\d\d)$" ) {      # mm/dd/yy
		$$dateRef = sprintf "%.2d-%.2d-%.2d", $3, $1, $2;
		return 1;
	}
	if ( $$dateRef =~ m"^(\d+)$" ) {                                   # day entered, use this month,year
		my ( $Year, $Month, $Day ) = ( $Now{year}, $Now{month}, $1 );
		if ( $Now{day} <= 3 and $Day > 21 ) {
			$Month--;
			$Month = 1, $Year-- if !$Month;
		}
		$$dateRef = sprintf "%.4d-%.2d-%.2d", $Year, $Month, $Day;
		return 1;
	}
	if ( $$dateRef =~ m"^(\d+)[/-](\d+)$" ) {                          # month/day entered,use this year
		$$dateRef = sprintf "%.4d-%.2d-%.2d", $Now{year}, $1, $2;
		return 1;
	}
	if ( $$dateRef =~ m"^([+-])(\d*)$" ) {                             # + or - from today
		my $inc = $2 ? $$dateRef : $$dateRef . '1';
		$$dateRef = sprintf "%.4d-%.2d-%.2d", Add_Delta_YMD( $Now{year}, $Now{month}, $Now{day}, 0, 0, $inc );
		return 1;
	}
	return 0;
}

# ---------------------------------------------------------------------------------------------
sub ValidDate {
	my ($dateRef) = @_;
	return 0 if !ReformatDate($dateRef);
	$$dateRef =~ m"(\d\d\d\d)-(\d\d)-(\d\d)";
	my ( $year, $month, $day ) = ( $1, $2, $3 );
	return Date::Calc::check_date( $year, $month, $day );
}

# ---------------------------------------------------------------------------------------------
sub ValidAmount {
	my ($amountRef) = @_;
	return FALSE if $$amountRef !~ m"^[ \t\$0-9.,\-\+]*$";
	$$amountRef =~ s"[ \t\$,]""g;
	return TRUE if $$amountRef !~ m"[\-\+]";
	my $amt = "0$$amountRef";
	$amt =~ s"^([0-9\.]+)"";
	my $tot = $1;
	while ($amt =~ s"^([\+\-])([0-9\.]+)"") {
		my ($op,$val) = ($1,$2);
		$val = -$val if $op eq '-';
		$tot += $val;
	}
	return FALSE if $amt;
	$$amountRef = $tot;
	return TRUE;
}

# ---------------------------------------------------------------------------------------------
sub ValidNumber {
	my ($numberRef) = @_;
	$$numberRef =~ s" ""g;
	return 1 if $$numberRef =~ m"^[ 0-9-]*$";
	return 0;
}

# ---------------------------------------------------------------------------------------------
sub FormatAmount {
	my ( $amount, $nopadding ) = @_;
	return $amount if $amount !~ m"^[0-9\.\-]+$";
	$amount = sprintf $columnFormats{amount}, $amount;
#	my $origLen = length($amount);
	while ( $amount =~ s"(\d)(\d\d\d)([,\.])"$1,$2$3" ) { $amount =~ s"^ ""; }
#	while ( length($amount) > $origLen and $amount =~ s"^ "" ) { }
	$amount =~ s" +$"" if $nopadding;
	$amount =~ s"^ +"" if $nopadding;
	return $amount;
}

sub FormatDate {
	my ($date) = @_;
	$date ||= '';
	return sprintf $columnFormats{date}, $date;
}

sub FormatCheck {
	my ($chk) = @_;
	my $fmtd = sprintf $columnFormats{check}, $chk;
	return ' ' x length($fmtd) if !$chk;
	return $fmtd;
}

sub FormatDesc {
	my ( $desc, $length ) = @_;
	$length = 30 if !$length;
	$desc = substr( $desc, 0, $length );
	return sprintf "%-${length}s  ", $desc;
}

# ---------------------------------------------------------------------------------------------
sub ManualStatementBalance {
	my ($statementRef) = @_;
	return FALSE if $statementRef->{statementdate};
	foreach my $field ( 'statementadditions1', 'statementadditions2', 'statementsubtractions1', 'statementsubtractions2' ) {
		return FALSE if !ValidAmount( \$statementRef->{$field} );
		$statementRef->{$field} ||= 0;
		return FALSE if $statementRef->{$field} != 0;
	}
	$statementRef->{statementendingbalance} = $statementRef->{beginningbalance} if $statementRef->{statementendingbalance} eq '';
	return TRUE;
}

# ---------------------------------------------------------------------------------------------
sub UpdateBalance {
	my ( $BalanceRef, $AccountType, $AmountRef, $TransactionType, $Factor ) = @_;
	$$AmountRef ||= '0';
	$Factor = 1 if !$Factor;
	$Factor *= -1 if $TransactionType !~ m"^$AccountType";
	print STDERR "        applying $$AmountRef to $$BalanceRef factor $Factor\n" if $showStatus and DEBUG;
	$$BalanceRef += ( $$AmountRef * $Factor );
}

# ---------------------------------------------------------------------------------------------
sub ApplyTransaction {    #apply to current account and statement as applicable
	my ( $transaction, $recalculateStatement, $reverse ) = @_;
	$recalculateStatement ||= '';
	$reverse              ||= '';
	my $statement = $Transactions{$transaction}{statementkey};
	my $account   = $Statements{$statement}{account};
	return if $statement eq $Accounts{$account}{recurringstatement};
	my $accounttype            = $Accounts{$account}{accounttype};
	my $transactiontype        = $Transactions{$transaction}{transactiontype};
	my $transactionamountfield = \$Transactions{$transaction}{amount};
	my $statementfield         = $AccountTypes{$accounttype}{transactiontypes}{$transactiontype};
	my $statementfieldref      = \$Statements{$statement}{$statementfield};
	my $accountbalancefield    = \$Accounts{$account}{balance};
	my $accountclearedfield    = \$Accounts{$account}{clearedbalance};
	Log( 'Applying transaction', $transaction, $transactiontype, $$transactionamountfield, 'reverse', $reverse );
	print STDERR "    applying transaction $transaction type $transactiontype stm $statement fld $statementfield acct $account type $accounttype reverse $reverse\n"
		if $showStatus and DEBUG;

	if ( $Transactions{$transaction}{cleared} ) {
		print STDERR "      updating statement\n" if $showStatus and DEBUG;
		Log( 'Updating statement', $Statements{$statement}{statementdate}, $statementfield, $$statementfieldref );
		UpdateBalance( $statementfieldref, $transactiontype, $transactionamountfield, $transactiontype, $reverse );
		Log( 'Updated statement', $Statements{$statement}{statementdate}, $statementfield, $$statementfieldref );
		print STDERR "      updating acctcleared\n" if $showStatus and DEBUG;
		Log( 'Updating account cleared', $account, $accounttype, $$accountclearedfield );
		UpdateBalance( $accountclearedfield, $accounttype, $transactionamountfield, $transactiontype, $reverse );
		Log( 'Updated account cleared', $account, $accounttype, $$accountclearedfield );
	}
	print STDERR "      updating acct\n" if $showStatus and DEBUG;
	Log( 'Updating account balance', $account, $accounttype, $$accountbalancefield );
	UpdateBalance( $accountbalancefield, $accounttype, $transactionamountfield, $transactiontype, $reverse );
	Log( 'Updated account balance', $account, $accounttype, $$accountbalancefield );
	RecalculateStatementBalances($statement) if $recalculateStatement;
	$Accounts{$account}{lastcheck} = $Transactions{$transaction}{checknumber} if $Transactions{$transaction}{checknumber} > $Accounts{$account}{lastcheck};
}

# ---------------------------------------------------------------------------------------------
sub RecalculateStatementBalances {
	my ($statement) = @_;
	print STDERR "  recalcing statement balances $statement ($Statements{$statement}{statementdate})\n" if $showStatus and DEBUG;
	foreach my $grp ( 'statement', 'transaction' ) {
		foreach my $fld ( 'additions1', 'additions2', 'subtractions1', 'subtractions2' ) {
			$Statements{$statement}{"${grp}${fld}"} ||= '0';
		}
		if ( $grp ne 'statement' or !ManualStatementBalance( $Statements{$statement} ) ) {
			$Statements{$statement}{"${grp}endingbalance"}
				= $Statements{$statement}{beginningbalance} + $Statements{$statement}{"${grp}additions1"} + $Statements{$statement}{"${grp}additions2"} - $Statements{$statement}{"${grp}subtractions1"} - $Statements{$statement}{"${grp}subtractions2"};
		}
		$Statements{$statement}{"${grp}endingbalance"} = 0 if abs( $Statements{$statement}{"${grp}endingbalance"} ) < 0.05;
		print STDERR "    ${grp}endingbalance - ", $Statements{$statement}{"${grp}endingbalance"}, "\n" if $showStatus and DEBUG;
	}
	foreach my $fld ( 'additions1', 'additions2', 'subtractions1', 'subtractions2', 'endingbalance' ) {
		$Statements{$statement}{"discrepancy$fld"} = $Statements{$statement}{"transaction$fld"} - $Statements{$statement}{"statement$fld"};
		$Statements{$statement}{"discrepancy$fld"} = 0 if abs( $Statements{$statement}{"discrepancy$fld"} ) < 0.005;
		print STDERR "    discrepancy$fld - ", $Statements{$statement}{"discrepancy$fld"}, "\n" if $showStatus and DEBUG;
	}
}

# ---------------------------------------------------------------------------------------------
sub RecalculateStatement {
	my ($statement) = @_;
	print STDERR "  recalcing statement $statement ($Statements{$statement}{statementdate})\n" if $showStatus and DEBUG;
	foreach my $field ( 'transactionadditions1', 'transactionadditions2', 'transactionsubtractions1', 'transactionsubtractions2' ) {
		$Statements{$statement}{$field} = 0;
	}
	my $date = ($isCurrent) ? 'Current statement' : "$Statements{$statement}{statementdate} statement";
	print STDERR "$date reset; stmtend=$Statements{$statement}{statementendingbalance}  trnend=$Statements{$statement}{transactionendingbalance}.\n"
		if $showStatus and DEBUG;
	foreach my $transaction ( ( tied %Transactions )->select_where("statementkey = '$statement'") ) {
		ApplyTransaction($transaction);
	}
	RecalculateStatementBalances($statement);
	print STDERR "    transactions re-applied; stmtend=$Statements{$statement}{statementendingbalance}  trnend=$Statements{$statement}{transactionendingbalance}.\n"
		if $showStatus and DEBUG;
}

# ---------------------------------------------------------------------------------------------
sub RecalculateBalances {
	my ($account) = @_;
	my @Accounts = ($account);
	@Accounts = sort keys %Accounts if !@Accounts;
	foreach my $account ( keys %Accounts ) {
		foreach my $field ( 'balance', 'clearedbalance', 'lastcheck' ) {
			$Accounts{$account}{$field} = 0;
		}
		print STDERR "\n$Accounts{$account}{accountname} reset.\n" if $showStatus and DEBUG;
		my $current   = $Accounts{$account}{currentstatement};
		my $recurring = $Accounts{$account}{recurringstatement};
		my $latest    = $Accounts{$account}{lateststatement};
		my %statementDates;
		foreach my $statement ( grep !m"^$current|$recurring$", ( tied %Statements )->select_where("account = '$account'") ) {
			$statementDates{ $Statements{$statement}{statementdate} } = $statement;
		}
		my @statements = ();
		foreach my $statementDate ( sort keys %statementDates ) {
			push @statements, ( $statementDates{$statementDate} );
		}
		my $oldestStatementDate             = "";
		my $oldestStatementBeginningBalance = 0;
		foreach my $statement ( @statements, $current ) {
			my $isCurrent = $statement eq $current;
			if ($isCurrent) {
				$Statements{$statement}{beginningbalance} = $latest ? $Statements{$latest}{statementendingbalance} : 0;
			}
			my $statementDate = $Statements{$statement}{statementdate};
			if ( $statementDate and $statementDate lt $oldestStatementDate ) {
				$oldestStatementDate             = $statementDate;
				$oldestStatementBeginningBalance = $Statements{$statement}{beginningbalance};
			}
			RecalculateStatement($statement);
		}

=pod
		if ($oldestStatementBeginningBalance) {
			foreach my $field ( 'balance', 'clearedbalance' ) {
				$Accounts{$account}{$field} += $oldestStatementBeginningBalance;
			}
		}
=cut

	}
}

# ---------------------------------------------------------------------------------------------
use vars qw ($fixAvailable);
$fixAvailable = FALSE;

sub FixStatements {
	$Session->{current}{message} = "No fix available.";
	return;
	$Session->{current}{message} = "";
	foreach my $statement ( keys %Statements ) {
		my $accountname   = $Accounts{ $Statements{$statement}{account} }{accountname};
		my $statementdate = $Statements{$statement}{statementdate};
		my $discrepancy   = $Statements{$statement}{discrepancyendingbalance};
		if ( $discrepancy ne '0' and abs($discrepancy) < 0.005 ) {
			$discrepancy = 0;
			$Statements{$statement}{discrepancyendingbalance} = $discrepancy;
		}
		my $additions2    = $Statements{$statement}{statementadditions2};
		my $subtractions2 = $Statements{$statement}{statementsubtractions2};
		my $hasproblem    = abs( $discrepancy - ( $subtractions2 * -2 ) ) < 0.005;
		if ( ( $discrepancy != 0 ) and ($hasproblem) and ( $additions2 == 0 ) ) {
			$Session->{current}{message} .= " $Accounts{$Statements{$statement}{account}}{accountname}:$Statements{$statement}{statementdate}";
			( $Statements{$statement}{statementadditions2}, $Statements{$statement}{statementsubtractions2} ) = ( $subtractions2, $additions2 );
			RecalculateStatementBalances($statement);
		}
	}
}

# ---------------------------------------------------------------------------------------------
sub EditRecord {
	my ($table) = @_;
	my $key = $Session->{current}{$table};
	return if !$key;
	$Session->{edit} = {} if !$Session->{edit};
	$Session->{edit}{$table} = {};
	foreach my $field ( @{ $Fields{$table} } ) {
		$Session->{edit}{$table}{$field} = $Records{$table}{$key}{$field} || '';
	}
	$Session->{edit}{$table}{key} = $key;
	return;
}

# ---------------------------------------------------------------------------------------------
sub NewRecord {
	my ($table) = @_;
	$Session->{edit} = {} if !$Session->{edit};
	foreach my $field ( @{ $Fields{$table} } ) {
		$Session->{edit}{$table}{$field} = ( $numericFields{$table}{$field} or $moneyFields{$table}{$field} ) ? 0 : '';
	}
	$Session->{edit}{$table}{key}             = 'new';
	$Session->{edit}{$table}{accounttype}     = $defaultAccountType if $table eq 'account';
	$Session->{edit}{$table}{transactiontype} = $defaultTransactionType if $table eq 'transaction';
	$Session->{edit}{$table}{increment}       = 'monthly' if $table eq 'transaction' and $isRecurring;
}

# ---------------------------------------------------------------------------------------------
sub InitialRecordFields {
	my ($token) = @_;
	foreach my $table (@Tables) {
		EditRecord( $table, $Session->{current}{$table} );
	}
	my $account   = $Session->{current}{account};
	my $statement = $Session->{current}{statement};
	if ( $account and $statement ) {
		$isCurrent   = $statement eq $Accounts{$account}{currentstatement};
		$isRecurring = $statement eq $Accounts{$account}{recurringstatement};
		$isLatest    = $statement eq $Accounts{$account}{lateststatement};
	}
}

# ---------------------------------------------------------------------------------------------
sub SortItem {
	my ( $sort, $date, $chk, $amt, $desc, $stm ) = @_;
	my $item = $date . $chk . $amt . $desc;    # sort eq 'date';
	$item = $desc . $date . $amt . $chk if $sort eq 'description';
	$item = $chk . $date . $amt . $desc if $sort eq 'checknumber';
	$item = $amt . $desc . $date . $chk if $sort eq 'amount';
	$item .= $stm if defined $stm;
	return $item;
}

# ---------------------------------------------------------------------------------------------
sub SortItem2 {
	my ( $sort, $date, $chk, $dr, $cr, $desc, $stm, $bal ) = @_;
	my $money = $dr . $cr;
	$money .= $bal if defined $bal;
	my $item = $date . $chk . $money . $desc;    # sort eq 'date';
	$item = $desc . $date . $money . $chk if $sort eq 'description';
	$item = $chk . $date . $money . $desc if $sort eq 'checknumber';
	$item = $money . $desc . $date . $chk if $sort eq 'amount';
	$item .= $stm if defined $stm;
	return $item;
}

# ---------------------------------------------------------------------------------------------
# End
1;
