package HTML;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	%colors
	$onload
	HTTPHeader
	HTMLStart
	HTMLEnd
	HTMLEnclosed
	HTMLNoEndTag
	HTMLStartEnd
	HTMLSingle
	HTMLSingleIndented
	HTMLBreaks
	Indent
	Outdent
	Indented
	HTMLHeader
	HTMLCloseWindow
	HTTPRedirect
	HTTPForbidden
	HTTPNotFound
	ResetFloat
	RadioButton
	InputField
	TableRow
	EditField
	ChoiceField
	HiddenField
	EditFormHeader
	EditFormFooter
	Select
	Button
	ClassAttribute
	IDAttribute
	StyleAttribute
	OnClick
	OnMouseOver
	OnMouseOut
	OnMouse
	Display
	LoadColors
	$htmlLevel $htmlIndent @htmlStack
);
use strict;
use warnings;
use Common;
use Multiuser;
use Data;

# ---------------------------------------------------------------------------------------------
# variables
use constant DEBUG => 0;
use vars qw (%colors $onload);

# ---------------------------------------------------------------------------------------------
# HTML functions
use vars qw ($htmlLevel $htmlIndent @htmlStack);
$htmlLevel  = 0;
$htmlIndent = "";
@htmlStack  = ();

sub HTTPHeader {
	print "HTTP/1.1 200 OK\nContent-Type: text/html\nCache-Control: no-cache\n\n";
}

sub HTMLStart {
	my ($code) = @_;
	my ($tag)  = $code =~ m"^(\S+)";
	push @htmlStack, ($tag);
	my $html = Indented("<$code>");
	Indent();
	return $html;
}

sub HTMLEnd {    # argument is allowed to document what is ended, but argument is ignored
	my ($code) = @_;
	my $tag = pop @htmlStack;
	return "<!-- no $code tag to end -->\n" if !$tag;
	my $comment = "";
	$comment = "<!-- $code end -->" if $code and $tag ne $code;
	Outdent();
	return Indented("</$tag>$comment");
}

sub HTMLEnclosed {
	my ( $code, @text ) = @_;
	my ($tag) = $code =~ m"^(\S+)";
	my $comment = "";
	$comment = "<!-- $tag enclosed -->" if DEBUG;
	return "<$code>" . join( "", @text ) . "</$tag>$comment";
}

sub HTMLNoEndTag {
	my ($code) = @_;
	return Indented("<$code>");
}

sub HTMLStartEnd {
	my ( $code, @text ) = @_;
	return Indented( HTMLEnclosed( $code, @text ) );
}

sub HTMLSingle {
	my ($code) = @_;
	return "<$code />";
}

sub HTMLSingleIndented {
	my ($code) = @_;
	return Indented("<$code />");
}

sub HTMLBreaks {
	my (@Text) = @_;
	map { s"\n"<br>\n"g; } @Text;
	return @Text;
}

sub Indent {
	my (@Text) = @_;
	$htmlLevel++;
	$htmlIndent = "  " x $htmlLevel;
}

sub Outdent {
	my (@Text) = @_;
	$htmlLevel--;
	$htmlIndent = "  " x $htmlLevel;
}

sub Indented {
	my (@Text) = @_;
	return "$htmlIndent" . join( "", @Text ) . "\n";
}

sub HTMLHeader {    # Headers are 'title','css','js'

	#	print '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">', "\n";
	#	print '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">', "\n";
	print '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">', "\n";

	#	print '<!DOCTYPE html">', "\n";
	print HTMLStart('html lang="en" xml:lang="en" xmlns="http://www.w3.org/1999/xhtml"');
	print HTMLStart('head');
	print HTMLSingleIndented('meta http-equiv="content-type" content="text/html;charset=utf-8"');
	print HTMLSingleIndented('meta http-equiv="X-UA-Compatible" content="IE=10"') if $UserAgent =~ m"IE";
	print HTMLStartEnd( 'title', "$Title V$Version" );
	print HTMLNoEndTag("link rel=StyleSheet href='/${Program}.css' type='text/css' media='screen'");
	print HTMLNoEndTag("link rel=StyleSheet href='/chrome.css' type='text/css' media='screen'")  if $UserAgent eq "Chrome";
	print HTMLNoEndTag("link rel=StyleSheet href='/firefox.css' type='text/css' media='screen'") if $UserAgent eq "Firefox";
	print HTMLNoEndTag("link rel=StyleSheet href='/ie9.css' type='text/css' media='screen'")     if $UserAgent eq "IE9";
	print HTMLNoEndTag("link rel=StyleSheet href='/ie8.css' type='text/css' media='screen'")     if $UserAgent eq "IE8";
	print HTMLNoEndTag("link rel=StyleSheet href='/ie7.css' type='text/css' media='screen'")     if $UserAgent eq "IE7";
	print HTMLStartEnd("script type='application/javascript' src='/${Program}.js'");
	print HTMLEnd('head');
}

sub HTMLCloseWindow {
	HTTPHeader();
	HTMLHeader();
	print HTMLEnclosed( 'script language=javascript', "window.open('', '_self', ''); window.close();" );
	print "Automatic window closing disabled, please close this window using Alt-F4.\n";
	print HTMLEnd('body');
	print HTMLEnd('html');
	return FALSE;
}

# ---------------------------------------------------------------------------------------------
sub HTTPRedirect {
	my ($location) = @_;
	$location ||= '';
	print "HTTP/1.1 302 Found\nContent-Type: text/html\nLocation: /$location\nContent-Location: /$location\nCache-Control: no-cache\nConnection: close\n\n";
	return FALSE;
}

sub HTTPForbidden {
	print "HTTP/1.1 403 Forbidden\nContent-Type: text/html\nCache-Control: no-cache\n\n<html><body>ERROR - That URL is not allowed.\n<br>\n</body></html>\n";
	return FALSE;
}

sub HTTPNotFound {
	my ($File) = @_;
	print "HTTP/1.1 404 Not Found\nContent-Type: text/html\nCache-Control: no-cache\n\n<html><body>ERROR - $File does not exist.\n<br>\n</body></html>\n";
	return FALSE;
}

# ---------------------------------------------------------------------------------------------
sub ResetFloat {
	print HTMLStartEnd( "div" . ClassAttribute('resetfloat'), ' ' );
}

sub RadioButton {
	my ( $table, $field, $value, $label ) = @_;
	$Session->{editable}{$table}{$field} = TRUE;
	my $checked = $value eq $Session->{edit}{$table}{$field} ? 'checked' : '';
	return HTMLSingle("input type=radio name=$field value='$value' $checked") . $label . HTMLSingle('br');
}

sub InputField {
	my ( $table, $field, $size, $readonly, $style, $tabindex ) = @_;
	$Session->{editable}{$table}{$field} = TRUE if !$readonly;
	my $value = $Session->{edit}{$table}{$field};
	my @styles = ( $style || '' );
	push @styles, ("text-align:right") if $numericFields{$table}{$field} or $moneyFields{$table}{$field};
	if ( $moneyFields{$table}{$field} ) { $size = 10 if !$size; ( $value = Amount($value) ) =~ s"^ *""; }
	push @styles, ( ClassBackground('readonly') ) if $readonly and !grep m"background-color", @styles;
	my ( $onMouseOver, $onMouseOut, $Tip ) = FieldHelp( $table, $field, $readonly );
	return HTMLSingle( "input type=text name=$field"
			. IDAttribute( $table, $field )
			. ValueAttribute( $Session->{edit}{$table}{$field} )
			. ClassAttribute('input')
			. OnMouse( $onMouseOver, $onMouseOut )
			. TabIndex($tabindex)
			. SizeAttribute($size)
			. ReadonlyAttribute($readonly)
			. StyleAttribute(@styles) )
		. $Tip;
}

sub FieldHelp {
	my ( $table, $field, $readonly ) = @_;
	return ( '', '', '' ) if $readonly or !$properties{fieldhelp} or !$table or !$field or !$FieldHelp{$table} or !$FieldHelp{$table}{$field};
	my $Tip         = HTMLEnclosed( "div" . IDAttribute( $table, $field, 'tip' ) . ClassAttribute('fieldhelp'), $FieldHelp{$table}{$field} );
	my $id          = $table . '_' . $field;
	my $onMouseOver = "showTip('$id');";
	my $onMouseOut  = "hideTip('$id');";
	return ( $onMouseOver, $onMouseOut, $Tip );
}

sub TableRow {
	my ( $label, $input ) = @_;
	print HTMLStartEnd( 'tr', HTMLEnclosed( 'th' . StyleAttribute("text-align:right"), $label ), HTMLEnclosed( 'td', $input ) );
}

sub EditField {
	my ( $table, $fieldlabel, $size, $readonly, $extra ) = @_;
	$extra = $extra ? "&nbsp;  $extra" : '';
	my $field = lc $fieldlabel;
	$field =~ s" ""g;
	TableRow( $fieldlabel, InputField( $table, $field, $size, $readonly ) . $extra );
}

sub ChoiceField {
	my ( $table, $fieldlabel, $readonly, $choices, $values ) = @_;
	( my $field = lc $fieldlabel ) =~ s" ""g;
	my $value;
	if ($table) {
		$value = $Session->{edit}{$table}{$field};
	} else {
		$value    = $readonly;
		$readonly = '';
	}
	if ($readonly) {
		$value = $$values{$value} if $values and $$values{$value};
		TableRow( $fieldlabel, HTMLSingle( "input type=text" . ValueAttribute($value) . ClassAttribute('input') . StyleAttribute( ClassBackground('readonly') ) . SizeAttribute( length($value) ) . ReadonlyAttribute($readonly) ) );
		return;
	}
	my ( $onMouseOver, $onMouseOut, $Tip ) = FieldHelp( $table, $field );
	$Session->{editable}{$table}{$field} = TRUE if $table;
	my $select = '';
	foreach my $choice (@$choices) {
		my $label = $values && $$values{$choice} ? $$values{$choice} : $choice;
		$value = $choice if !defined $value or $value eq '';
		my $selected = $choice eq $value ? 'selected' : '';
		$select .= HTMLEnclosed( "option value='$choice' $selected", $label );
	}
	$select = HTMLEnclosed( "select name='$field'" . IDAttribute( $table, $field ) . ClassAttribute('choice') . OnMouse( $onMouseOver, $onMouseOut ), $select );
	TableRow( $fieldlabel, $select );
}

sub HiddenField {
	my ( $field, $value ) = @_;
	print HTMLNoEndTag("input type=hidden name=$field value=\"$value\"");
}

sub EditFormHeader {
	my ( $table, $key ) = @_;
	my $edit = "edit$table";
	print HTMLStart( 'div' . ClassAttribute("$edit content") );
	print HTMLStart("form method='POST' id='$edit' action=''");    # name='$edit'
	HiddenField( 'action', '' );
	HiddenField( 'form',   $table );
	HiddenField( 'key',    $key ) if $key;
	print HTMLStart( 'table' . ClassAttribute( $edit . 'table' ) );    # width=100%
	print HTMLStart( 'tbody' . ClassAttribute( $edit . 'table' ) );    # width=100%
}

sub EditFormFooter {
	print HTMLEnd('tbody');
	print HTMLEnd('table');
	print HTMLEnd('form');
	print HTMLEnd('div');
}

sub Select {
	my ( $label, $class, $table, $key ) = @_;
	print STDERR "No current $table.\n" if !exists $Session->{current}{$table} and $showStatus;
	return HTMLEnclosed( 'div' . ClassAttribute( $class . ' selected' ), $label ) if $key eq $Session->{current}{$table};
	return HTMLEnclosed( 'div' . ClassAttribute( $class . ' clickable' ) . OnMouse( ClassBackground( 'clickable', TRUE ), ClassBackground( $class, TRUE ) ) . OnClick( 'Select', $table, $key ), $label );
}

sub Button {
	my ( $label, $class, @function ) = @_;
	my $AccessKey = '';
	if ( $label =~ s"&(.)"$1" ) {
		my $ShortcutKey = uc $1;
		$AccessKey = "accesskey='$ShortcutKey'";
		$label .= " (Alt-$ShortcutKey)";
	}
	my $color = ClassColor($class);
	return HTMLSingle(
		"input type='submit' $AccessKey"
			. ValueAttribute($label) . ClassAttribute( $class . ' clickable' ) . OnMouse( "this.style.background='#000';this.style.color='$color';", "this.style.background='$color';this.style.color='#000';" ) . OnClick(@function)
	);
}

sub TabIndex {
	my ($index) = @_;
	return " tabindex=$index" if $index;
	return '';
}

sub IDAttribute {
	my (@id) = @_;
	return '' if !@id or !$id[0];
	return " id='" . join( '_', @id ) . "'";
}

sub ValueAttribute {
	my ($value) = @_;
	return " value=\"$value\"";
}

sub SizeAttribute {
	my ($size) = @_;
	return " size=$size" if $size;
	return '';
}

sub ReadonlyAttribute {
	my ($readonly) = @_;
	return " readonly='readonly'" if $readonly;
	return '';
}

sub ClassAttribute {
	my ($class) = @_;
	return " class='$class'" if $class;
	return '';
}

sub StyleAttribute {
	my (@styles) = @_;
	return '' if !@styles;
	map { s"^ *""; s" *$""; s";$""; } @styles;
	my $style = join( ';', @styles ) . ';';
	return " style='$style'";
}

sub OnClick {
	my ( $function, $parameter, @moreparameters ) = @_;
	return '' if !$function;
	my $onClick = " onClick='$function(\"$parameter\"";
	foreach my $moreparameter (@moreparameters) { $onClick .= ",\"$moreparameter\""; }
	$onClick .= ");'";
	return $onClick;
}

sub OnMouseOver {
	my ($onMouseOver) = @_;
	$onMouseOver =~ s"^ *"";
	$onMouseOver =~ s" *$"";
	return '' if !$onMouseOver;
	return " onMouseOver=\"$onMouseOver\"";
}

sub OnMouseOut {
	my ($onMouseOut) = @_;
	$onMouseOut =~ s"^ *"";
	$onMouseOut =~ s" *$"";
	return '' if !$onMouseOut;
	return " onMouseOut=\"$onMouseOut\"";
}

sub OnMouse {
	my ( $onMouseOver, $onMouseOut ) = @_;
	return OnMouseOver($onMouseOver) . OnMouseOut($onMouseOut);
}

sub Display {
	my ($field) = @_;
	$field =~ s" *:\d+$"";
	$field =~ s" "&nbsp;"g;
	return "<code>$field</code>";
}

sub ClassColor {
	my ($colorClass) = @_;
	my $color = 'gray';
	foreach my $class ( split m" +", $colorClass ) {
		next if !$colors{$class}{'background-color'};
		$color = $colors{$class}{'background-color'};
		last;
	}
	print STDERR "Class '$colorClass' - no color set!\n" if $color eq 'gray' and -t STDERR;
	return $color;
}

sub ClassBackground {
	my ( $colorClass, $isJs ) = @_;
	my $color = ClassColor($colorClass);
	return "this.style.background='$color';" if $isJs;
	return "background-color:$color;";
}

sub Pointer {
	my ($pointer) = @_;
	$pointer ||= 'default';
	return "document.body.style.cursor='$pointer';";
}

sub LoadColors {
	open my $CSS, "<", "$Program.css";
	chomp( my @CSS = <$CSS> );
	close $CSS;
	my $css = join( '', @CSS );
	$css =~ s"/\*[^/]*\*/""g;
	@CSS = split m"\}", $css;
	foreach my $CSS (@CSS) {
		next if $CSS !~ m"^(.*)\{(.*)";
		my ( $classes, $styles ) = ( $1, $2 );
		$classes =~ s"\s""g;
		foreach my $class ( split m"[,]+", $classes ) {
			next if $class !~ s"^\."";
			foreach my $style ( split m";", $styles ) {
				next if $style !~ m"(.*):(.*)";
				my ( $var, $value ) = ( $1, $2 );
				$var =~ s"\s""g;
				$value =~ s"\s""g;
				$colors{$class}{$var} = $value if $value =~ m"^#[0-9a-f]+$"i;
			}
		}
	}
}

# ---------------------------------------------------------------------------------------------
# End
1;
__END__
