package Process;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	ProcessPage
);
use constant DEBUG => 0;
use strict;
use warnings;
use DateIncrement;
use Common;
use Multiuser;
use Data;
use HTML;
use Authenticate;
use Queries;
use Reports;
use Pages;

# ---------------------------------------------------------------------------------------------
# ProcessPage
sub ProcessPage {
	my ($object) = @_;
	$onload                      = "";
	$Session->{current}{message} = $GlobalMessage;    # normally blank
	$GlobalMessage               = '';
	SetError("");
	$Session->{current}{action} = "";
	$Session->{current}{adminmode} ||= FALSE;
	$Session->{current}{account}   = '' if $Session->{current}{account}   and $Session->{current}{account} ne 'new'   and !$Accounts{ $Session->{current}{account} };
	$Session->{current}{statement} = '' if $Session->{current}{statement} and $Session->{current}{statement} ne 'new' and !$Statements{ $Session->{current}{statement} };
	$Session->{current}{transaction} = ''
		if $Session->{current}{transaction}
		and $Session->{current}{transaction} ne 'new'
		and !$Transactions{ $Session->{current}{transaction} };
	$Session->{current}{detail} = '' if $Session->{current}{detail} and $Session->{current}{detail} ne 'new' and !$Details{ $Session->{current}{detail} };
	$input{view}   ||= "";
	$input{action} ||= "";
	$input{form}   ||= "";
	my $action = $input{action};
	my $form   = $input{form};
	my $view   = $input{view} || $Session->{current}{view};
	return MassSelect()              if $action eq 'massselect';
	return HTTPRedirect('configure') if $view eq 'configure';
	return HTTPRedirect('users')     if $view eq 'users';
	$viewAccounts     = $view eq 'accounts';
	$viewStatements   = $view eq 'statements';
	$viewTransactions = $view eq 'transactions';
	$viewQueries      = $view eq 'queries';
	$viewReports      = $view eq 'reports';
	###
	$isCurrent = $isRecurring = $isLatest = FALSE;
	if ( $Session->{current}{statement} ) {
		$isCurrent   = $Session->{current}{statement} eq $Accounts{ $Session->{current}{account} }{currentstatement};
		$isRecurring = $Session->{current}{statement} eq $Accounts{ $Session->{current}{account} }{recurringstatement};
		$isLatest    = $Session->{current}{statement} eq $Accounts{ $Session->{current}{account} }{lateststatement};
	}
	###
	$isNewAccount = $isNewStatement = $isNewTransaction = $isExistingAccount = $isExistingStatement = $isExistingTransaction = FALSE;
	$isNewAccount          = TRUE if $Session->{current}{account}     and $Session->{current}{account} eq 'new';
	$isNewStatement        = TRUE if $Session->{current}{statement}   and $Session->{current}{statement} eq 'new';
	$isNewTransaction      = TRUE if $Session->{current}{transaction} and $Session->{current}{transaction} eq 'new';
	$isExistingAccount     = TRUE if $Session->{current}{account}     and $Session->{current}{account} ne 'new';
	$isExistingStatement   = TRUE if $Session->{current}{statement}   and $Session->{current}{statement} ne 'new';
	$isExistingTransaction = TRUE if $Session->{current}{transaction} and $Session->{current}{transaction} ne 'new';
	###
	if ( $view eq 'archived' ) {
		$Session->{current}{showarchived} = !$Session->{current}{showarchived};
		$Session->{current}{account}      = "";
		$Session->{current}{statement}    = "";
		$Session->{current}{transaction}  = "";
		$Session->{current}{detail}       = "";
		return;
	}
	if ( $view eq 'accountinfo' ) {
		my $account = $Session->{current}{account};
		return if !$account;
		my $Acct = $Accounts{$account};
		my $info = "$$Acct{accountname}\n";
		if ( $$Acct{accounttype} eq 'c' ) {
			$info .= "Account: $$Acct{accountnumber}\n"     if $$Acct{accountnumber};
			$info .= "Expiration: $$Acct{expirationdate}\n" if $$Acct{expirationdate};
			$info .= "CSC: $$Acct{securitycode}\n"          if $$Acct{securitycode};
		} else {
			$info .= "Account: $$Acct{accountnumber}\n" if $$Acct{accountnumber};
			$info .= "Routing: $$Acct{routingnumber}\n" if $$Acct{routingnumber};
			$info .= "CardNumber: $$Acct{cardnumber} "  if $$Acct{cardnumber};
			$info .= "Expires $$Acct{expirationdate} "  if $$Acct{expirationdate};
			$info .= "CSC $$Acct{securitycode}"         if $$Acct{securitycode};
			$info .= "\n";
		}
		$Session->{current}{message} = $info;
		return;
	}
	if ( $view eq 'recalc' ) {
		my $account = $Session->{current}{account};
		return if !$account;
		RecalculateBalances($account);
		EditRecord('account');
		EditRecord('statement');
		EditRecord('transaction');
		EditRecord('detail');
		$Session->{current}{message} = "$Accounts{$account}{accountname} balances recalculated.";
		return;
	}
	if ( $view eq 'recalcstatement' ) {
		my $account = $Session->{current}{account};
		return if !$account;
		my $statement = $Session->{current}{statement};
		return if !$statement;
		RecalculateStatement($statement);
		EditRecord('account');
		EditRecord('statement');
		EditRecord('transaction');
		EditRecord('detail');
		$Session->{current}{message} = "$Accounts{$account}{accountname} statement $Statements{$statement}{statementdate} balances recalculated.";
		return;
	}
	SelectField( $input{field}, $input{value} ) if $action eq 'select';
	InputRecord($form) if $form;
	foreach my $property ( 'view', 'action' ) {
		$Session->{current}{$property} = $input{$property} if $input{$property};
	}
	$Session->{current}{deleteclicked} = "" if !$action or $action !~ "^delete";
	if ( $form eq 'account' ) {
		return SaveAccount()     if $action eq 'save';
		return DeleteAccount()   if $action eq 'delete';
		return ArchiveAccount()  if $action eq 'archive';
		return ActivateAccount() if $action eq 'activate';
		return;
	}
	if ( $form eq 'statement' ) {
		return SaveStatement()          if $action eq 'save';
		return FinalizeStatement()      if $action eq 'finalize';
		return UnfinalizeStatement()    if $action eq 'unfinalize';
		return PrintStatement()         if $action eq 'print';
		return PrintUncleared()         if $action eq 'printuncleared';
		return PrintUncleared(TRUE)     if $action eq 'printunclearedfix';
		return FixStatements()          if $action eq 'fix';
		return PrintFullReport()        if $action eq 'fullreport';
		return PrintBalanceReport()     if $action eq 'balancereport';
		return PrintCreditUtilization() if $action eq 'creditutilization';
		return;
	}
	if ( $form eq 'transaction' ) {
		return SaveTransaction()   if $action eq 'save';
		return DeleteTransaction() if $action eq 'delete';
		return;
	}
	if ( $form eq 'detail' ) {
		return SaveDetail()   if $action eq 'save';
		return DeleteDetail() if $action eq 'delete';
		return;
	}
	if ( $form eq 'query' ) {
		QueryTransactions($action);
		return               if $action eq 'query';
		return PrintQuery()  if $action eq 'print';
		return ExportQuery() if $action eq 'export';
		return;
	}
}

# ---------------------------------------------------------------------------------------------
sub MassSelect {
	my %transactions;
	foreach my $input ( sort keys %input ) {
		$transactions{$1} = $2 if $input =~ m"^(\d+)\.(clear|unclear|post)$";
	}
	if ( !%transactions ) {
		$Session->{current}{message} = "Nothing selected!";
		return;
	}
	my @messages = ();
	my %actions;
	foreach my $transaction ( keys %transactions ) {
		my $action = $transactions{$transaction};
		$actions{$action}++;
		ClearTransaction($transaction)   if $action eq 'clear';
		UnclearTransaction($transaction) if $action eq 'unclear';
		PostTransaction($transaction)    if $action eq 'post';
		push @messages, ( $Session->{current}{message} ) if $Session->{current}{message};
	}
	if ( @messages and scalar(@messages) <= 3 ) {
		$Session->{current}{message} = join( '  ', @messages );
	} else {
		$Session->{current}{message} = '';
		my %labels = ( 'clear' => 'Cleared', 'unclear' => 'Uncleared', 'post' => 'Posted' );
		foreach my $action ( sort keys %actions ) {
			$Session->{current}{message} .= "$actions{$action} transactions $labels{$action}.  ";
		}
	}
}

# ---------------------------------------------------------------------------------------------
sub InputRecord {
	my ($table) = @_;
	if ( $table eq 'query' ) {
		foreach my $field ( keys %{ $Session->{edit}{query} } ) {
			$Session->{edit}{$table}{$field} = $input{$field} if exists $input{$field};
		}
		ValidDate( \$Session->{edit}{$table}{date} )   if $Session->{edit}{$table}{date};
		ValidDate( \$Session->{edit}{$table}{todate} ) if $Session->{edit}{$table}{todate};
		$Session->{edit}{$table}{amount}   = Amount( $Session->{edit}{$table}{amount} )   if $Session->{edit}{$table}{amount};
		$Session->{edit}{$table}{toamount} = Amount( $Session->{edit}{$table}{toamount} ) if $Session->{edit}{$table}{toamount};
		return;
	}
	foreach my $field ( 'key', @{ $Fields{$table} } ) {
		$Session->{edit}{$table}{$field} = $input{$field} if exists $input{$field};
	}
}

# ---------------------------------------------------------------------------------------------
sub SetMode {
	my ($mode) = @_;
	$Session->{current}{$mode} = $Session->{current}{$mode} ? FALSE : TRUE;
	$Session->{current}{$mode} = FALSE if $mode eq 'adminmode' and !isAdmin();
}

# ---------------------------------------------------------------------------------------------
sub SelectField {
	my ( $table, $key ) = @_;
	return DoReport($key)              if $table eq 'report';
	return SelectTransactionSort($key) if $table eq 'transactionsort';
	return ClearTransaction($key)      if $table eq 'clear';
	return UnclearTransaction($key)    if $table eq 'unclear';
	return PostTransaction($key)       if $table eq 'post';
	return SetMode($key)               if $table eq 'mode';
	return DeleteTransaction($key)     if $table eq 'deletetransaction';
	return UndeleteTransaction()       if $table eq 'undeletetransaction';
	return SelectQueryResult($key)     if $table eq 'queryresult';
	return                             if $Session->{current}{$table} eq $key;
	$Session->{current}{adminmode} = FALSE;

	if ( $table eq 'account' ) {
		$Session->{current}{view} = 'accounts' if !$Session->{current}{view};
		$Session->{current}{view} = 'accounts' if $key eq 'new';
		$Session->{current}{account}     = $key;
		$Session->{current}{statement}   = "";
		$Session->{current}{transaction} = "";
		$Session->{current}{detail}      = "";
		if   ( $key eq 'new' ) { $isNewAccount = TRUE;  $isExistingAccount = FALSE; }
		else                   { $isNewAccount = FALSE; $isExistingAccount = TRUE; }
		$isNewStatement        = FALSE;
		$isExistingStatement   = FALSE;
		$isNewTransaction      = FALSE;
		$isExistingTransaction = FALSE;
		$onload                = 'account_accountname' if $Session->{current}{view} eq 'accounts';
	} elsif ( $table eq 'statement' ) {
		$Session->{current}{view}        = 'transactions';
		$Session->{current}{statement}   = $key;
		$Session->{current}{transaction} = "";
		$Session->{current}{detail}      = "";
		$isCurrent                       = $key eq $Accounts{ $Session->{current}{account} }{currentstatement};
		$isRecurring                     = $key eq $Accounts{ $Session->{current}{account} }{recurringstatement};
		$isLatest                        = $key eq $Accounts{ $Session->{current}{account} }{lateststatement};
		if   ( $key eq 'new' ) { $isNewStatement = TRUE;  $isExistingStatement = FALSE; }
		else                   { $isNewStatement = FALSE; $isExistingStatement = TRUE; }
		$isNewTransaction      = FALSE;
		$isExistingTransaction = FALSE;
		$onload                = 'statement_statementdate';
	} elsif ( $table eq 'transaction' ) {
		$Session->{current}{view}        = 'transactions';
		$Session->{current}{transaction} = $key;
		$Session->{current}{detail}      = "";
		if   ( $key eq 'new' ) { $isNewTransaction = TRUE;  $isExistingTransaction = FALSE; }
		else                   { $isNewTransaction = FALSE; $isExistingTransaction = TRUE; }
		$onload = 'transaction_date';
	} elsif ( $table eq 'detail' ) {
		$Session->{current}{view}   = 'transactions';
		$Session->{current}{detail} = $key;
		$onload                     = 'detail_amount';
	}
	if ( $key eq 'new' ) {
		NewRecord($table);
	} else {
		EditRecord($table);
		if ( $table eq 'account' ) {
			$Session->{current}{statement} = $Accounts{ $Session->{current}{account} }{currentstatement};
			$isCurrent                     = TRUE;
			$isRecurring                   = FALSE;
			EditRecord('statement');
		}
	}
}

# ---------------------------------------------------------------------------------------------
sub DoReport {
	my ($report) = @_;
	return PrintStatement()         if $report eq 'print';
	return PrintUncleared()         if $report eq 'printuncleared';
	return PrintRecurring()         if $report eq 'printrecurring';
	return PrintFullReport()        if $report eq 'fullreport';
	return PrintUncleared(TRUE)     if $report eq 'printunclearedfix';
	return PrintBalanceReport()     if $report eq 'balancereport';
	return PrintCreditUtilization() if $report eq 'creditutilization';
}

# ---------------------------------------------------------------------------------------------
sub SaveAccount {    # existence of all input fields is assumed
	$onload = 'account_accountname';
	my $account = $Session->{current}{account};
	$onload = 'account_accountname',    return SetError("Account Name is required.")                                                                    if !$Session->{edit}{account}{accountname};
	$onload = 'account_accounttype',    return SetError("Account Type is required.")                                                                    if !$Session->{edit}{account}{accounttype};
	$onload = 'account_accountnumber',  return SetError("'$Session->{edit}{account}{accountnumber}' is not a valid account number.")                    if !ValidNumber( \$Session->{edit}{account}{accountnumber} );
	$onload = 'account_routingnumber',  return SetError("'$Session->{edit}{account}{routingnumber}' is not a valid bank routing number.")               if !ValidNumber( \$Session->{edit}{account}{routingnumber} );
	$onload = 'account_routingnumber',  return SetError("Routing Number must be 9 digits.")                                                             if $Session->{edit}{account}{routingnumber} and length( $Session->{edit}{account}{routingnumber} ) != 9;
	$onload = 'account_cardnumber',     return SetError("'$Session->{edit}{account}{cardnumber}' is not a valid card number.")                          if !ValidNumber( \$Session->{edit}{account}{cardnumber} );
	$onload = 'account_expirationdate', return SetError("Expiration date '$Session->{edit}{account}{expirationdate}' is not valid, format is 'mm/yy'.") if $Session->{edit}{account}{expirationdate} !~ m"^(\d\d/\d\d)?$";
	$onload = 'account_securitycode',   return SetError("Security code '$Session->{edit}{account}{securitycode}' is not valid, must be 3 digits.")      if $Session->{edit}{account}{securitycode} !~ m"^(\d\d\d)?$";
	if ( $account ne 'new' and $Session->{edit}{account}{accounttype} ne $Accounts{$account}{accounttype} ) {
		$Session->{edit}{account}{accounttype} = $Accounts{$account}{accounttype};
		$onload = 'account_accounttype';
		return SetError("Cannot change account type on an existing account.");
	}
	if ( $account ne 'new' ) {
		foreach my $field ( keys %{ $Session->{editable}{account} } ) {
			$Accounts{$account}{$field} = $Session->{edit}{account}{$field};
		}
		$Session->{current}{message} = "Bank Account $Accounts{$account}{accountname} updated.";
		if ( $Accounts{$account}{lateststatement} == $Accounts{$account}{currentstatement} ) {
			my $latestStatement     = 0;
			my $latestStatementDate = 0;
			foreach my $statement ( ( tied %Statements )->select_where("account = '$account'") ) {
				next if $statement eq $Accounts{$account}{currentstatement} or $statement eq $Accounts{$account}{recurringstatement};
				my $date = $Statements{$statement}{statementdate};
				next if $date lt $latestStatementDate;
				$latestStatementDate = $date;
				$latestStatement     = $statement;
			}
			if ( $latestStatement == $Accounts{$account}{currentstatement} ) {
				SetError("Correction made to selection of Last Statement.");
			} else {
				$Accounts{$account}{lateststatement} = $latestStatement;
				SetError("Correction made to selection of Last Statement.");
			}
		}
		return;
	}
	$account = AddAccountEdit();
	return if !$account;
	$Session->{current}{account} = $account;
	$Session->{current}{message} = "Bank Account $Accounts{$account}{accountname} created.";
	EditRecord('account');
	$Accounts{$account}{currentstatement}   = AddStatementNew( $account, 0 );
	$Accounts{$account}{recurringstatement} = AddStatementNew( $account, 0 );
	$Accounts{$account}{lateststatement}    = 0;
	$onload                                 = 'account_accountname';
}

# ---------------------------------------------------------------------------------------------
sub DeleteAccount {
	if ( !$Session->{current}{deleteclicked} ) {
		$Session->{current}{deleteclicked} = 1;
		return SetError("Click delete again to delete Bank Account $Accounts{$Session->{current}{account}}{accountname}.");
	}
	my $AccountName = $Accounts{ $Session->{current}{account} }{accountname};
	delete $Accounts{ $Session->{current}{account} };
	$Session->{current}{view}        = "";
	$Session->{current}{account}     = "";
	$Session->{current}{statement}   = "";
	$Session->{current}{transaction} = "";
	$Session->{current}{detail}      = "";
	$Session->{current}{message}     = "Bank Account $AccountName deleted.";
}

# ---------------------------------------------------------------------------------------------
sub ArchiveAccount {
	$Accounts{ $Session->{current}{account} }{archived} = 1;
	$Session->{current}{message}                        = "Bank Account $Accounts{$Session->{current}{account}}{accountname} archived.";
	$Session->{current}{account}                        = "";
	$Session->{current}{view}                           = "";
}

# ---------------------------------------------------------------------------------------------
sub ActivateAccount {
	$Accounts{ $Session->{current}{account} }{archived} = 0;
	$Session->{current}{message}                        = "Bank Account $Accounts{$Session->{current}{account}}{accountname} re-activated.";
	$Session->{current}{account}                        = "";
	$Session->{current}{view}                           = "";
}

# ---------------------------------------------------------------------------------------------
sub SaveStatement {
	$onload = ManualStatementBalance( $Session->{edit}{statement} ) ? 'statement_statementendingbalance' : 'statement_statementadditions1';
	my $account   = $Session->{current}{account};
	my $statement = $Session->{current}{statement};
	if ( $Session->{edit}{statement}{statementdate} ne $Statements{$statement}{statementdate} and $statement eq $Accounts{$account}{currentstatement} ) {    # if Admin mode could be statement other than current
		my $date = $Session->{edit}{statement}{statementdate};
		if ( !$date ) {
			$date = 0;
		} else {
			if ( !ValidDate( \$date ) ) {
				$onload = 'statement_statementdate';
				return SetError("Invalid Statement Date.");
			}
			$Session->{edit}{statement}{statementdate} = $date;
			my $latestStatement = $Accounts{$account}{lateststatement};
			if ($latestStatement) {
				my $latestStatementDate = $Statements{$latestStatement}{statementdate};
				if ( $date le $latestStatementDate ) {
					$onload = 'statement_statementdate';
					return SetError("Statement Date must be later than previous statement date $latestStatementDate.");
				}
			}
		}
	}
	my @checkFields = ( 'statementadditions1', 'statementadditions2', 'statementsubtractions1', 'statementsubtractions2' );
	push @checkFields, ('statementendingbalance') if ManualStatementBalance( $Session->{edit}{statement} );
	foreach my $field (@checkFields) {
		my $fld = \$Session->{edit}{statement}{$field};
		if ( !ValidAmount($fld) ) {
			$onload = 'statement_$field';
			return SetError("Invalid amount '$$fld' in field '$field'.");
		}
	}
	foreach my $field ( keys %{ $Session->{editable}{statement} } ) {
		$Statements{$statement}{$field} = $Session->{edit}{statement}{$field};
	}
	RecalculateStatementBalances($statement);
	EditRecord('statement');
	$Session->{current}{message} = "Statement updated.";
	return 1;
}

# ---------------------------------------------------------------------------------------------
sub FinalizeStatement {
	my $account   = $Session->{current}{account};
	my $statement = $Session->{current}{statement};
	my $date      = $Session->{edit}{statement}{statementdate};
	return SetError("Date is required to finalize statement.")        if !$date;
	return                                                            if !SaveStatement();
	return SetError("There's a balance discrepancy, can't finalize.") if $Statements{$statement}{discrepancyendingbalance};
	my $oldCurrent = $statement;
	$statement = AddStatementNew( $account, $Statements{$oldCurrent}{statementendingbalance} );
	$Accounts{$account}{lateststatement}  = $oldCurrent;
	$Accounts{$account}{currentstatement} = $statement;
	$Session->{current}{statement}        = $statement;
	$Session->{current}{transaction}      = '';
	$Session->{current}{detail}           = '';
	my $clearedCount   = 0;
	my $unclearedCount = 0;

	foreach my $transaction ( ( tied %Transactions )->select_where("statementkey = '$oldCurrent'") ) {
		if ( $Transactions{$transaction}{cleared} ) {
			$clearedCount++;
			next;
		}
		$Transactions{$transaction}{statementkey} = $statement;
		$unclearedCount++;
	}
	$Session->{current}{message} = "Statement $date finalized; $clearedCount transactions cleared, $unclearedCount transactions open.";
	$Session->{current}{'statement'} = '';
	SelectField( 'statement', $statement );
}

# ---------------------------------------------------------------------------------------------
sub UnfinalizeStatement {
	my $account    = $Session->{current}{account};
	my $newCurrent = $Accounts{$account}{lateststatement};
	return SetError("No statement exists to unfinalize.\n") if !$newCurrent;
	my $oldCurrent = $Accounts{$account}{currentstatement};
	return SetError("Error!  No current statement.\n") if !$oldCurrent;
	my @alreadyCleared = ( tied %Transactions )->select_where("statementkey = '$oldCurrent' and cleared = '1'");
	return SetError("Cannot unfinalize latest statement when there are current transactions already cleared.") if @alreadyCleared;
	foreach my $transaction ( ( tied %Transactions )->select_where("statementkey = '$oldCurrent'") ) {
		$Transactions{$transaction}{statementkey} = $newCurrent;
	}
	delete $Statements{$oldCurrent};
	$Accounts{$account}{currentstatement} = $newCurrent;
	my $latestStatement     = 0;
	my $latestStatementDate = 0;
	foreach my $statement ( ( tied %Statements )->select_where("account = '$account'") ) {
		next if $statement eq $Accounts{$account}{currentstatement} or $statement eq $Accounts{$account}{recurringstatement};
		my $date = $Statements{$statement}{statementdate};
		next if $date lt $latestStatementDate;
		$latestStatementDate = $date;
		$latestStatement     = $statement;
	}
	$Accounts{$account}{lateststatement} = $latestStatement;
	$Session->{current}{message}         = "Statement $Session->{edit}{statement}{statementdate} unfinalized.";
	$Session->{current}{'statement'}     = '';
	SelectField( 'statement', $newCurrent );
}

# ---------------------------------------------------------------------------------------------
sub SaveTransaction {
	$onload = 'transaction_amount';
	my $transaction = $Session->{current}{transaction};
	my $statement   = $Session->{current}{statement};
	my $account     = $Session->{current}{account};
	$Session->{edit}{transaction}{description} = $TransactionTypes{ $Session->{edit}{transaction}{transactiontype} } if !$Session->{edit}{transaction}{description};
	if ( !$Session->{edit}{transaction}{description} ) {
		$Session->{edit}{transaction}{description} = $Transactions{$transaction}{description} if $transaction ne 'new';
		$onload = 'transaction_description';
		return SetError("Description is required.");
	}
	if ( !$isCurrent and !$isRecurring and $Transactions{$transaction}{transactiontype} ne $Session->{edit}{transaction}{transactiontype} ) {
		$Session->{edit}{transaction}{transactiontype} = $Transactions{$transaction}{transactiontype};
		$onload = 'transaction_transactiontype';
		return SetError("Cannot change Transaction Type on a closed transaction.");
	}
	return SetError("Invalid amount '$Session->{edit}{transaction}{amount}'.") if !ValidAmount( \$Session->{edit}{transaction}{amount} );
	$Session->{edit}{transaction}{checknumber} = $Accounts{$account}{lastcheck} + 1 if $Session->{edit}{transaction}{checknumber} eq '+';
	return SetError("Invalid check number.") if !ValidNumber( \$Session->{edit}{transaction}{checknumber} );
	return SetError("Check number is required for a Check transaction.") if $Session->{edit}{transaction}{transactiontype} eq 'c3' and !$Session->{edit}{transaction}{checknumber} and !$isRecurring;
	if ( !ValidDate( \$Session->{edit}{transaction}{date} ) ) {
		$onload = 'transaction_date';
		return SetError("Invalid date.");
	}
	foreach my $field ( sort keys %{ $Session->{editable}{transaction} } ) {
		next if $Session->{edit}{transaction}{$field} !~ m"\?";
		$onload = 'transaction_$field';
		return SetError("$field cannot contain questionmarks!");
	}
	$Session->{edit}{transaction}{checknumber} = 0 if !$Session->{edit}{transaction}{checknumber};
	if ( $transaction ne 'new' ) {
		my $updateBalances = FALSE;
		if ($isCurrent) {
			$updateBalances = TRUE if $Transactions{$transaction}{amount} != $Session->{edit}{transaction}{amount};
			$updateBalances = TRUE if $Transactions{$transaction}{transactiontype} ne $Session->{edit}{transaction}{transactiontype};
			$updateBalances = TRUE if $Transactions{$transaction}{checknumber} ne $Session->{edit}{transaction}{checknumber};
		}
		ApplyTransaction( $transaction, '', REVERSE ) if $updateBalances;
		foreach my $field ( sort keys %{ $Session->{editable}{transaction} } ) {
			$Transactions{$transaction}{$field} = $Session->{edit}{transaction}{$field};
		}
		ApplyTransaction( $transaction, RECALCULATE ) if $updateBalances;
		EditRecord('transactions');
		EditRecord('statement');
		$Session->{current}{message} = $Transactions{$transaction}{cleared} ? 'Cleared transaction updated.' : 'Transaction updated.';
		$Session->{current}{message} = "Cleared transaction modified." if !$isCurrent;
		$Session->{current}{message} = "Recurring transaction updated." if $isRecurring;
		CheckDuplicateTransactions();
		return;
	}
	$transaction = AddTransactionEdit($statement);
	return if !$transaction;
	$Session->{current}{transaction} = $transaction;
	$Session->{edit}{transaction}{key} = $transaction;
	ApplyTransaction( $transaction, RECALCULATE );
	EditRecord('transactions');
	EditRecord('statement');
	$Session->{current}{message} = "Transaction added.";
	CheckDuplicateTransactions();
}

# ---------------------------------------------------------------------------------------------
sub CheckDuplicateTransactions {
	my $statement              = $Session->{current}{statement};
	my $currentTransaction     = $Session->{current}{transaction};
	my $currentAmount          = $Transactions{$currentTransaction}{amount};
	my $currentDate            = $Transactions{$currentTransaction}{date};
	my $currentDescription     = $Transactions{$currentTransaction}{description};
	my $currentDescriptionUC   = uc $currentDescription;
	my $currentCleared         = !!$Transactions{$currentTransaction}{cleared};         # ensure 0 or 1
	my $currentTranType        = $Transactions{$currentTransaction}{transactiontype};
	my $isIdentical_msg        = '';
	my $seemsIdentical_msg     = '';
	my $matchingAmountDate_msg = '';
	my $matchingAmountDesc_msg = '';
	my $similarAmountDesc_msg  = '';
	my $matchingDateDesc_msg   = '';
	my $similarDateDesc_msg    = '';
	foreach my $transaction ( ( tied %Transactions )->select_where("statementkey = '$statement' order by date desc") ) {
		next if $transaction eq $currentTransaction or $Transactions{$transaction}{transactiontype} ne $currentTranType;
		my $amount             = $Transactions{$transaction}{amount};
		my $date               = $Transactions{$transaction}{date};
		my $description        = $Transactions{$transaction}{description};
		my $descriptionUC      = uc $description;
		my $cleared            = !!$Transactions{$transaction}{cleared};
		my $amountMatches      = $amount eq $currentAmount;
		my $dateMatches        = $date eq $currentDate;
		my $descriptionMatches = $descriptionUC eq $currentDescriptionUC;
		my $descriptionSimilar = ( index( $descriptionUC, $currentDescriptionUC ) != -1 ) || ( index( $currentDescriptionUC, $descriptionUC ) != -1 );
		my $clearedMessage     = $cleared == $currentCleared ? '' : "  That transaction is " . ( $cleared ? " cleared." : "open." );

		if ( $amountMatches and $dateMatches and $descriptionMatches ) {
			$isIdentical_msg = "<br>This transaction is identical to one on the same date." . $clearedMessage if !$isIdentical_msg;
		} elsif ( $amountMatches and $dateMatches and $descriptionSimilar ) {
			$seemsIdentical_msg = "<br>This transaction appears to be identical to '$description' on the same date." . $clearedMessage if !$seemsIdentical_msg;
		} elsif ( $amountMatches and $dateMatches ) {
			$matchingAmountDate_msg = "<br>This transaction has the same amount and date as '$description'." . $clearedMessage if !$matchingAmountDate_msg;
		} elsif ( $amountMatches and $descriptionMatches ) {
			$matchingAmountDesc_msg = "<br>This transaction has the same amount and description as one on $date." . $clearedMessage if !$matchingAmountDesc_msg;
		} elsif ( $amountMatches and $descriptionSimilar ) {
			$similarAmountDesc_msg = "<br>This transaction has the same amount and a description similar to '$description' on $date." . $clearedMessage if !$similarAmountDesc_msg;
		} elsif ( $dateMatches and $descriptionMatches ) {
			$matchingDateDesc_msg = "<br>This transaction has the same date and description as another one." . $clearedMessage if !$matchingDateDesc_msg;
		} elsif ( $dateMatches and $descriptionSimilar ) {
			$similarDateDesc_msg = "<br>This transaction has the same date as and a description similar to another one." . $clearedMessage if !$similarDateDesc_msg;
		}
	}
	if ($isIdentical_msg) {
		$Session->{current}{message} .= $isIdentical_msg;
	} elsif ($seemsIdentical_msg) {
		$Session->{current}{message} .= $seemsIdentical_msg;
	} else {
		$Session->{current}{message} .= $matchingAmountDate_msg . $matchingAmountDesc_msg . $matchingDateDesc_msg;
		$Session->{current}{message} .= $similarAmountDesc_msg if !$matchingAmountDesc_msg;
		$Session->{current}{message} .= $similarDateDesc_msg if !$matchingDateDesc_msg;
	}
}

# ---------------------------------------------------------------------------------------------
sub DeleteTransaction {
	my ($transaction) = @_;
	$transaction = $Session->{current}{transaction} if !$transaction;
	if ( $transaction eq 'new' ) {
		$Session->{current}{transaction} = '';
		return;
	}
	ApplyTransaction( $transaction, RECALCULATE, REVERSE );
	foreach my $field (@deletedTransactionFields) {
		$Session->{deleted}{$field} = $Transactions{$transaction}{$field};
	}
	delete $Transactions{$transaction};
	$Session->{current}{transaction} = "";
	$Session->{current}{detail}      = "";
	EditRecord('statement');
	$Session->{current}{message} = "Transaction deleted.";
}

# ---------------------------------------------------------------------------------------------
sub UndeleteTransaction {
	$Session->{current}{view}        = 'transactions';
	$Session->{current}{transaction} = 'new';
	$Session->{current}{detail}      = "";
	$onload                          = 'transaction_date';
	NewRecord('transaction');
	$Session->{deleted}{transactiontype} = $DefaultTransactionTypes{ substr( $Session->{deleted}{transactiontype}, 0, 1 ) }
		if !$AccountTypes{ $Accounts{ $Session->{current}{account} }{accounttype} }{transactiontypes}{ $Session->{deleted} };
	foreach my $field (@deletedTransactionFields) {
		$Session->{edit}{transaction}{$field} = $Session->{deleted}{$field};
	}
	SaveTransaction();
	$Session->{current}{message} =~ s" added\." undeleted.";
}

# ---------------------------------------------------------------------------------------------
sub SelectTransactionSort {
	my ($sort) = @_;
	$Session->{current}{transactionsort} = $sort;
}

# ---------------------------------------------------------------------------------------------
sub ClearTransaction {
	my ($transaction) = @_;
	ApplyTransaction( $transaction, '', REVERSE );
	$Transactions{$transaction}{cleared} = 1;
	ApplyTransaction( $transaction, RECALCULATE );
	EditRecord('statement');
	EditRecord('transaction') if $transaction eq $Session->{current}{transaction};
}

# ---------------------------------------------------------------------------------------------
sub UnclearTransaction {
	my ($transaction) = @_;
	ApplyTransaction( $transaction, '', REVERSE );
	$Transactions{$transaction}{cleared} = 0;
	ApplyTransaction( $transaction, RECALCULATE );
	EditRecord('statement');
	EditRecord('transaction') if $transaction eq $Session->{current}{transaction};
}

# ---------------------------------------------------------------------------------------------
sub PostTransaction {
	my ($transaction) = @_;
	my $newDate = IncrementDate( $Transactions{$transaction}{date}, $Transactions{$transaction}{increment} || 'monthly' );
	return if $Session->{current}{error};
	my $account        = $Session->{current}{account};
	my $statement      = $Accounts{$account}{currentstatement};
	my $newTransaction = AddTransactionRecord( $statement, $transaction );
	return if !$newTransaction;
	my $msg = "Transaction posted on $Transactions{$transaction}{date} for \$$Transactions{$transaction}{amount} to $Transactions{$transaction}{description}.";
	ApplyTransaction( $newTransaction, RECALCULATE );
	$Transactions{$transaction}{date} = $newDate;

	foreach my $detail ( ( tied %Details )->select_where("transactionkey = '$transaction'") ) {
		my $rows = $addDetail->execute( $newTransaction, $Details{$detail}{description}, $Details{$detail}{amount} );
	}
	$Session->{current}{transaction} = $transaction;
	$Session->{current}{detail}      = "";
	EditRecord('transaction');
	$Session->{current}{message} = $msg;
}

# ---------------------------------------------------------------------------------------------
sub SaveDetail {
	$onload = 'detail_amount';
	my $detail      = $Session->{current}{detail};
	my $transaction = $Session->{current}{transaction};
	return SetError("Invalid amount.") if !ValidAmount( \$Session->{edit}{detail}{amount} );
	if ( $detail ne 'new' ) {
		foreach my $field ( keys %{ $Session->{editable}{detail} } ) {
			$Details{$detail}{$field} = $Session->{edit}{detail}{$field};
		}
		EditRecord('transaction');
		$Session->{current}{message} = "Detail updated.";
		return;
	}
	$detail = AddDetailEdit($transaction);
	return if !$detail;
	$Session->{current}{detail} = $detail;
	$Session->{edit}{detail}{key} = $detail;
	EditRecord('detail');
	EditRecord('transaction');
	$Session->{current}{message} = "Detail added.";
}

# ---------------------------------------------------------------------------------------------
sub DeleteDetail {
	my $detail = $Session->{current}{detail};
	if ( $detail eq 'new' ) {
		$Session->{current}{detail} = '';
		return;
	}
	delete $Details{$detail};
	$Session->{current}{detail}  = "";
	$Session->{current}{message} = "Detail deleted.";
}

# ---------------------------------------------------------------------------------------------
sub SelectQueryResult {
	my ($transaction) = @_;
	my $statement     = $Transactions{$transaction}{statementkey};
	my $account       = $Statements{$statement}{account};
	$Session->{current}{view}        = 'transactions';
	$Session->{current}{account}     = '';
	$Session->{current}{statement}   = '';
	$Session->{current}{transaction} = '';
	$Session->{current}{detail}      = '';
	SelectField( 'account',     $account );
	SelectField( 'statement',   $statement );
	SelectField( 'transaction', $transaction );
}

# ---------------------------------------------------------------------------------------------
sub SetError {
	my ($Error) = @_;
	$Session->{current}{error} = $Error;
	return 0;
}

# ---------------------------------------------------------------------------------------------
# End
1;
