package Queries;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	ShowQueryForm
	ShowQueryResults
	PrintQuery
	ExportQuery
	QueryTransactions
);
use strict;
use warnings;
use File::Basename;
use PDFUtil;
use Common;
use Multiuser;
use Data;
use HTML;

# ---------------------------------------------------------------------------------------------
sub ShowQueryForm {
	foreach my $field ( 'account', 'date', 'todate', 'amount', 'toamount', 'description', 'checknumber', 'tochecknumber', 'transactionsort' ) {
		$Session->{edit}{query}{$field} ||= '';
	}
	$Session->{edit}{query}{account} = $Session->{current}{account} if !$Session->{edit}{query}{account};
	EditFormHeader('query');
	print HTMLStartEnd( "tr", HTMLEnclosed( "th colspan=2 class='queryheader'", "Query Transactions" ) );
	my %showAccounts;
	foreach my $account ( keys %Accounts ) {
		$showAccounts{ $Accounts{$account}{accountname} } = $account;
	}
	my %accountList = ( 'all' => 'All Accounts' );
	my @accountList = ('all');
	foreach my $accountName ( sort { uc($a) cmp uc($b) } keys %showAccounts ) {
		my $account = $showAccounts{$accountName};
		$accountList{$account} = $accountName;
		push @accountList, ($account);
	}
	ChoiceField( '', 'Account', $Session->{edit}{query}{account}, \@accountList, \%accountList );
	EditField( 'query', 'Date' );
	EditField( 'query', 'To date' );
	EditField( 'query', 'Amount' );
	EditField( 'query', 'To amount' );
	EditField( 'query', 'Description' );
	EditField( 'query', 'Check Number' );
	EditField( 'query', 'To Check Number' );
	my %sortList;
	my @sortList;

	foreach my $label ( 'Date', 'Description', 'Check Number', 'Amount' ) {
		my $sort = lc $label;
		$sort =~ s" ""g;
		push @sortList, ($sort);
		$sortList{$sort} = $label;
	}
	ChoiceField( '', 'Transaction Sort', $Session->{edit}{query}{transactionsort}, \@sortList, \%sortList );
	print HTMLStart('tr');
	print HTMLStart('td colspan=2 align=right');
	print Indented( Button( 'Query',  'editquerybutton', 'Submit', 'editquery', 'query' ) );
	print Indented( Button( 'Print',  'editquerybutton', 'Submit', 'editquery', 'print', TRUE ) );
	print Indented( Button( 'Export', 'editquerybutton', 'Submit', 'editquery', 'export' ) );
	print HTMLEnd('td');
	print HTMLEnd('tr');
	EditFormFooter();
}

# ---------------------------------------------------------------------------------------------
sub ShowQueryResults {
	return if $Session->{current}{action} ne 'query';
	$Session->{current}{queryresult} = '';    # for HTML::Select
	print HTMLStart( 'div' . ClassAttribute('content queryresults') );
	print HTMLStartEnd( 'div' . ClassAttribute('queryresultsheader'), scalar(@queryResults) . " Query Results" );
	print HTMLStart( 'div' . ClassAttribute('content queryresultslist') );
	foreach my $result (@queryResults) {
		$result =~ s" (\d+)$"";
		my $transaction = $1;
		$result =~ s"  $"";
		print Select( Display($result) . '<br>', 'querylistitem', 'queryresult', $transaction );
	}
	print "No transactions match criteria.\n" if !@queryResults;
	print HTMLEnd('div');
	print HTMLEnd('div');
}

# ---------------------------------------------------------------------------------------------
sub PrintQuery {
	return if $Session->{current}{action} ne 'print';
	if ( !@queryResults ) {
		$Session->{current}{error} = 'No transactions match criteria.';
		return;
	}
	my $ColumnHeader
		= SortItem( $Session->{edit}{query}{transactionsort}, $columnHeaders{date}, $columnHeaders{checknumber}, $columnHeaders{amount}, $columnHeaders{description}, $columnHeaders{statementdate} );
	my $account = $Session->{edit}{query}{account} || 'all';
	$account = $Accounts{$account}{accountname} if $account ne 'all';
	my $SelectHeader = "Account: $account";
	if ( $Session->{edit}{query}{date} ) {
		$SelectHeader .= "   Date: $Session->{edit}{query}{date}";
		$SelectHeader .= " to $Session->{edit}{query}{todate}" if $Session->{edit}{query}{todate};
	}
	if ( $Session->{edit}{query}{amount} ) {
		$SelectHeader .= '   Amount: ' . sprintf( "%.2f", $Session->{edit}{query}{amount} );
		$SelectHeader .= ' to ' . sprintf( "%.2f", $Session->{edit}{query}{toamount} ) if $Session->{edit}{query}{toamount};
	}
	if ( $Session->{edit}{query}{checknumber} ) {
		$SelectHeader .= "   CheckNumber: $Session->{edit}{query}{checknumber}";
		$SelectHeader .= " to $Session->{edit}{query}{tochecknumber}" if $Session->{edit}{query}{tochecknumber};
	}
	$SelectHeader .= "   Description: '$Session->{edit}{query}{description}'" if $Session->{edit}{query}{description};
	my @SelectHeaders = ();
	while ( length($SelectHeader) > 80 ) {
		my $line = substr( $SelectHeader, 0, 80, '' );
		my $keep = $line =~ s"\s*(\S*:[^:]*)$"" ? $1 : '';
		$SelectHeader = "$keep$SelectHeader";
		push @SelectHeaders, ($line);
	}
	push @SelectHeaders, ($SelectHeader);
	my $PDF = PDFOpen( "$ReportsDir/Query", LANDSCAPE, 'Transaction Query', $ColumnHeader, 'Query Criteria', @SelectHeaders );
	if ( !ref($PDF) ) {
		$Session->{current}{error} = $PDF;
		return;
	}
	foreach my $result (@queryResults) {
		$result =~ s" (\d+)$"";
		$result =~ s" 0 "   "g;
		PDFPrintLine( $PDF, $result );
	}
	PDFPrintLine( $PDF, '' );
	PDFPrintLine( $PDF, SortItem( $Session->{edit}{query}{transactionsort}, FormatDate(''), FormatCheck(0), FormatAmount($queryTotal), FormatDesc('Transaction Total'), FormatDate('') ) );
	PDFClose($PDF);
	$Session->{current}{message} = "Query printout '$$PDF{Filename}' created.";
	$DownloadFile = $$PDF{Filename};
}

# ---------------------------------------------------------------------------------------------
sub ExportQuery {
	return if $Session->{current}{action} ne 'export';
	if ( !@queryResults ) {
		$Session->{current}{error} = 'No transactions match criteria.';
		return;
	}
	$DownloadContent
		= join( ",", split( m" +", SortItem( $Session->{edit}{query}{transactionsort}, $columnHeaders{date}, $columnHeaders{checknumber}, $columnHeaders{amount}, $columnHeaders{description}, $columnHeaders{statementdate} ) ), 'Account' ) . "\n";
	my $account = $Session->{edit}{query}{account} || 'all';
	$account = $Accounts{$account}{accountname} if $account ne 'all';
	my $Select = "Account: $account";
	if ( $Session->{edit}{query}{date} ) {
		$Select .= "   Date: $Session->{edit}{query}{date}";
		$Select .= " to $Session->{edit}{query}{todate}" if $Session->{edit}{query}{todate};
	}
	if ( $Session->{edit}{query}{amount} ) {
		$Select .= '   Amount: ' . sprintf( "%.2f", $Session->{edit}{query}{amount} );
		$Select .= ' to ' . sprintf( "%.2f", $Session->{edit}{query}{toamount} ) if $Session->{edit}{query}{toamount};
	}
	if ( $Session->{edit}{query}{checknumber} ) {
		$Select .= "   CheckNumber: $Session->{edit}{query}{checknumber}";
		$Select .= " to $Session->{edit}{query}{tochecknumber}" if $Session->{edit}{query}{tochecknumber};
	}
	$Select .= "   Description: '$Session->{edit}{query}{description}'" if $Session->{edit}{query}{description};
	foreach my $result (@queryResults) {
		$DownloadContent .= "$result\n";
	}
	$DownloadFilename = 'Export.csv';
	$Session->{current}{message} = "Query export complete.";
}

# ---------------------------------------------------------------------------------------------
sub QueryTransactions {
	my ($action) = @_;
	@queryResults = ();
	$queryTotal   = 0;
	my @Accounts = ( $Session->{edit}{query}{account} );
	@Accounts = keys %Accounts if !$Session->{edit}{query}{account} or $Session->{edit}{query}{account} eq 'all';
	my $fromdate = $Session->{edit}{query}{date};
	my $todate   = $Session->{edit}{query}{todate};
	if ( $todate and !$fromdate ) {
		$fromdate = '1900-01-01';
		$Session->{edit}{query}{date} = $fromdate;
	}
	if ($fromdate) {
		ValidDate( \$todate ) if $fromdate;
		$Session->{edit}{query}{todate} = $todate;
	}
	my $fromamount = $Session->{edit}{query}{amount};
	my $toamount   = $Session->{edit}{query}{toamount};
	$toamount = $fromamount if !$toamount;
	my $description     = $Session->{edit}{query}{description};
	my $fromchecknumber = $Session->{edit}{query}{checknumber};
	my $tochecknumber   = $Session->{edit}{query}{tochecknumber};
	$tochecknumber = $fromchecknumber if !$tochecknumber;
	my @sql;
	push @sql, ( "date >= '$fromdate'",               "date <= '$todate'" )               if $fromdate;
	push @sql, ( "amount >= '$fromamount'",           "amount <= '$toamount'" )           if $fromamount;
	push @sql, ( "checknumber >= '$fromchecknumber'", "checknumber <= '$tochecknumber'" ) if $fromchecknumber;

	foreach my $account (@Accounts) {
		my $accountname = $account eq $Session->{edit}{query}{account} ? '' : sprintf "%-10s", substr( $Accounts{$account}{accountname}, 0, 30 );
		foreach my $statement ( ( tied %Statements )->select_where("account = '$account'") ) {
			next if $statement eq $Accounts{$account}{recurringstatement};
			foreach my $transaction ( ( tied %Transactions )->select_where( join( ' and ', "statementkey = '$statement'", @sql ) ) ) {
				my $date = $Transactions{$transaction}{date};
				my $chk  = $Transactions{$transaction}{checknumber};
				my $amt  = $Transactions{$transaction}{amount};
				my $desc = $Transactions{$transaction}{description};
				my $stm  = $Statements{ $Transactions{$transaction}{statementkey} }{statementdate};
				$stm = "(current)" if !$stm and $Transactions{$transaction}{cleared};
				$stm = "(not clr)" if !$stm;
				next if $description and !DescriptionMatch( $description, $transaction );
				UpdateBalance( \$queryTotal, $Accounts{$account}{accounttype}, \$amt, $Transactions{$transaction}{transactiontype} );
				$date = FormatDate($date);
				$chk  = FormatCheck($chk);
				$amt  = FormatAmount($amt);
				$desc = FormatDesc($desc);
				$stm  = FormatDate($stm);

				if ( $action eq 'export' ) {
					map { s" *$""; } ( $date, $stm );
					push @queryResults, ("$Q$date$Q,$Q$chk$Q,$Q$amt$Q,$Q$desc$Q,$Q$stm$Q,$Q$accountname$Q");
				} else {
					push @queryResults, ( SortItem( $Session->{edit}{query}{transactionsort}, $date, $chk, $amt, $desc, $stm ) . " $accountname $transaction" );
				}
			}
		}
	}
	if ( $action ne 'export' ) {
		@queryResults = sort { uc($a) cmp uc($b) } @queryResults;
		@queryResults = reverse @queryResults if $Session->{edit}{query}{transactionsort} eq 'date';
	}
	$Session->{current}{message} = scalar(@queryResults) . " matching transactions found.  Transaction total is " . sprintf( "%.2f", $queryTotal ) . ".";
}

# ---------------------------------------------------------------------------------------------
sub DescriptionMatch {
	my ( $description, $transaction ) = @_;
	return TRUE if $Transactions{$transaction}{description} =~ m"$description"i;
	return TRUE if $Transactions{$transaction}{description2} =~ m"$description"i;
	return TRUE if $Transactions{$transaction}{description3} =~ m"$description"i;
	foreach my $detail ( ( tied %Details )->select_where("transactionkey = '$transaction'") ) {
		return TRUE if $Details{$detail}{description} =~ m"$description"i;
	}
	return FALSE;
}

# ---------------------------------------------------------------------------------------------
# End
1;
