package Reports;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);    # Inherit from Exporter
@EXPORT = qw(
	PrintStatement
	PrintUncleared
	PrintRecurring
	PrintFullReport
	PrintBalanceReport
	PrintCreditUtilization
);
use strict;
use warnings;
use File::Basename;
use PDFUtil;
use Common;
use Multiuser;
use Data;
use HTML;

# ---------------------------------------------------------------------------------------------
sub PrintStatement {
	my ( $statement, $PDF ) = @_;
	$statement = $Session->{current}{statement} if !$statement;
	return if !$statement;
	my $account       = $Statements{$statement}{account};
	my $accountname   = $Accounts{$account}{accountname};
	my $accounttype   = $Accounts{$account}{accounttype};
	my $statementdate = $Statements{$statement}{statementdate};
	$statementdate = 'Current' if $statement eq $Accounts{$account}{currentstatement};
	my $sort         = $Session->{current}{transactionsort};
	my $header       = "$accountname - $statementdate Statement";
	my $ColumnHeader = SortItem2( $sort, $columnHeaders{date}, $columnHeaders{checknumber}, $columnHeaders{credits}, $columnHeaders{debits}, $columnHeaders{description} );
	my $closePDF     = FALSE;

	if ( !$PDF ) {
		$PDF = PDFOpen( "$ReportsDir/Statement-$accountname-$statementdate", PORTRAIT, $header, $ColumnHeader );
		$closePDF = TRUE;
	} else {
		PDFHeaders( $PDF, $header, $ColumnHeader );
	}
	if ( !ref($PDF) ) {
		$Session->{current}{error} = $PDF;
		return;
	}
	my @items  = ();
	my $totalD = 0;
	my $totalC = 0;
	foreach my $transaction ( ( tied %Transactions )->select_where("statementkey = '$statement' and cleared = '1'") ) {
		my $isD  = $Transactions{$transaction}{transactiontype} =~ m"^d";
		my $isC  = !$isD;
		my $date = $Transactions{$transaction}{date};
		my $chk  = $Transactions{$transaction}{checknumber};
		my $amt  = $Transactions{$transaction}{amount};
		$totalD += $amt if $isD;
		$totalC += $amt if $isC;
		my $desc = $Transactions{$transaction}{description};
		my $stm  = $Statements{ $Transactions{$transaction}{statementkey} }{statementdate};
		$date = FormatDate($date);
		$chk  = FormatCheck($chk);
		$amt  = FormatAmount($amt);
		$desc = FormatDesc($desc);
		$stm  = FormatDate($stm);
		my ( $d, $c ) = ( $amt, $emptyFields{amount} );
		( $c, $d ) = ( $d, $c ) if $isC;
		push @items, ( SortItem2( $sort, $date, $chk, $d, $c, $desc ) );
	}
	foreach my $item ( sort @items ) {
		$item =~ s" (\d+)$"";
		$item =~ s"&nbsp;" "g;
		$item =~ s"<[^>]*>""g;
		$item =~ s" 0 "   "g;
		PDFPrintLine( $PDF, $item );
	}
	PDFPrintLine( $PDF, '' ), PDFPrintLine( $PDF, "Nothing to print!" ), PDFPrintLine( $PDF, '' ) if !@items;
	PDFDrawLine($PDF);
	PDFPrintLine( $PDF, SortItem2( $sort, $emptyFields{date}, $emptyFields{check}, FormatAmount($totalD), FormatAmount($totalC), FormatDesc('Total') ) );
	PDFPrintLine( $PDF, '' );
	PDFPrintLine( $PDF, '' );
	my $format = '%20s %12.2f %12.2f %12.2f\n';
	my ( $Begin, @Rows ) = @{ $StatementTotalTypes{$accounttype} };
	my ( $field, $label ) = split m"=", $Begin;
	PDFPrintLine( $PDF, sprintf( "%20s %12s %12s %12s\n", ' ', 'Statement', 'Transactions', 'Discrepancy' ), '', TRUE );
	PDFPrintLine( $PDF, sprintf "%20s %12.2f\n", $label, $Statements{$statement}{$field} );

	foreach my $Row (@Rows) {
		my ( $field, $label ) = split m"=", $Row;
		PDFPrintLine( $PDF, sprintf $format, $label, $Statements{$statement}{"statement$field"}, $Statements{$statement}{"transaction$field"}, $Statements{$statement}{"discrepancy$field"} );
	}
	SetAndClosePDF($PDF) if $closePDF;
	$Session->{current}{message} = "$header report created.";
}

# ---------------------------------------------------------------------------------------------
sub PrintUncleared {
	my ( $fixAccount, $PDF ) = @_;
	my $account = $Session->{current}{account};
	return if !$account;
	my $statement = $Accounts{$account}{currentstatement};
	return if !$statement;
	my $accountname           = $Accounts{$account}{accountname};
	my $accounttype           = $Accounts{$account}{accounttype};
	my $sort                  = $Session->{current}{transactionsort};
	my $header                = "$accountname - Uncleared Transactions";
	my $ColumnHeader          = SortItem2( $sort, $columnHeaders{date}, $columnHeaders{checknumber}, $columnHeaders{credits}, $columnHeaders{debits}, $columnHeaders{description}, undef, $columnHeaders{balance} );
	my $accountclearedbalance = $Accounts{$account}{clearedbalance};
	my $accountbalance        = $Accounts{$account}{balance};
	my $statementbalance      = $Statements{$statement}{statementendingbalance};
	my $balance               = $statementbalance;
	my @items                 = ();
	my %items;
	my $totalD = 0;
	my $totalC = 0;

	foreach my $transaction ( ( tied %Transactions )->select_where("statementkey = '$statement' and cleared = '0'") ) {
		my $isD  = $Transactions{$transaction}{transactiontype} =~ m"^d";
		my $isC  = !$isD;
		my $date = $Transactions{$transaction}{date};
		my $chk  = $Transactions{$transaction}{checknumber};
		my $amt  = $Transactions{$transaction}{amount};
		my $desc = $Transactions{$transaction}{description};
		my $stm  = $Transactions{$transaction}{cleared};
		$totalD += $amt if $isD;
		$totalC += $amt if $isC;
		my $increment = substr( $Transactions{$transaction}{transactiontype}, 0, 1 ) eq $accounttype ? $amt : -$amt;
		$date = FormatDate($date);
		$chk  = FormatCheck($chk);
		$amt  = FormatAmount($amt);
		$desc = FormatDesc($desc);
		my ( $d, $c ) = ( $amt, $emptyFields{amount} );
		( $c, $d ) = ( $d, $c ) if $isC;
		my $item = SortItem2( $sort, $date, $chk, $d, $c, $desc );
		push @items, ($item);
		$items{$item} = [ $increment, $sort, $date, $chk, $d, $c, $desc ];
	}
	if ( !@items and !$fixAccount ) {
		$Session->{current}{message} = "No uncleared transactions found." if !$PDF;
		return;
	}
	my $closePDF = FALSE;
	if ( !$PDF ) {
		$PDF = PDFOpen( "$ReportsDir/Uncleared-$accountname", LANDSCAPE, $header, $ColumnHeader );
		$closePDF = TRUE;
	} else {
		PDFHeaders( $PDF, $header, $ColumnHeader );
	}
	if ( !ref($PDF) ) {
		$Session->{current}{error} = $PDF;
		return;
	}
	PDFPrintLine($PDF);
	if ( abs( $accountclearedbalance - $statementbalance ) > 0.05 ) {
		PDFPrintLine( $PDF, SortItem2( $sort, $emptyFields{date}, $emptyFields{check}, PrintCrDrBalance( $accounttype, $accountclearedbalance ), FormatDesc('Account Cleared Balance'), undef, $emptyFields{balance} ) );
		if ($fixAccount) {
			$Accounts{$account}{clearedbalance} = $statementbalance;
			PDFPrintLine( $PDF, 'Account cleared balance set to Statement Ending Balance' );
			PDFPrintLine($PDF);
		}
	}
	PDFPrintLine( $PDF, SortItem2( $sort, $emptyFields{date}, $emptyFields{check}, PrintCrDrBalance( $accounttype, $statementbalance ), FormatDesc('Statement Ending Balance'), undef, $emptyFields{balance} ) );
	PDFPrintLine($PDF);
	PDFDrawLine($PDF);
	foreach my $item ( sort @items ) {
		$item =~ s" (\d+)$"";
		$item =~ s"&nbsp;" "g;
		$item =~ s"<[^>]*>""g;
		$item =~ s" 0 "   "g;
		my ( $increment, @details ) = @{ $items{$item} };
		$balance += $increment;
		PDFPrintLine( $PDF, SortItem2( @details, undef, FormatAmount($balance) ) );
	}
	PDFDrawLine($PDF);
	PDFPrintLine( $PDF, SortItem2( $sort, $emptyFields{date}, $emptyFields{check}, FormatAmount($totalD), FormatAmount($totalC), FormatDesc('Total'), undef, $emptyFields{balance} ) );
	PDFDrawLine($PDF);
	PDFPrintLine($PDF);
	PDFPrintLine($PDF);
	PDFPrintLine( $PDF, SortItem2( $sort, $emptyFields{date}, $emptyFields{check}, PrintCrDrBalance( $accounttype, $balance ), FormatDesc('Calculated Balance'), undef, $emptyFields{balance} ), '', TRUE );
	if ( abs( $accountbalance - $balance ) > 0.05 ) {    # this should be !=, but I'm getting a != even when 'eq' is true!
		PDFPrintLine( $PDF, SortItem2( $sort, $emptyFields{date}, $emptyFields{check}, PrintCrDrBalance( $accounttype, $accountbalance ), FormatDesc('Account Balance') ) );
		if ($fixAccount) {
			$Accounts{$account}{balance} = $balance;
			PDFPrintLine( $PDF, 'Account balance set to Calculated Balance' );
		}
	}
	SetAndClosePDF($PDF) if $closePDF;
	$Session->{current}{message} = "$header report created.";
}

# ---------------------------------------------------------------------------------------------
sub PrintRecurring {
	my ($PDF) = @_;
	my $account = $Session->{current}{account};
	return if !$account;
	my $statement = $Accounts{$account}{recurringstatement};
	return if !$statement;
	my $accountname  = $Accounts{$account}{accountname};
	my $accounttype  = $Accounts{$account}{accounttype};
	my $sort         = $Session->{current}{transactionsort};
	my $header       = "$accountname - Recurring Transactions";
	my $ColumnHeader = SortItem2( $sort, $columnHeaders{date}, $columnHeaders{checknumber}, $columnHeaders{credits}, $columnHeaders{debits}, $columnHeaders{description} );
	my @items        = ();
	foreach my $transaction ( ( tied %Transactions )->select_where("statementkey = '$statement'") ) {
		my $isD  = $Transactions{$transaction}{transactiontype} =~ m"^d";
		my $isC  = !$isD;
		my $date = $Transactions{$transaction}{date};
		my $chk  = $Transactions{$transaction}{checknumber};
		my $amt  = $Transactions{$transaction}{amount};
		my $desc = $Transactions{$transaction}{description};
		my $stm  = $Transactions{$transaction}{cleared};
		$date = FormatDate($date);
		$chk  = FormatCheck($chk);
		$amt  = FormatAmount($amt);
		$desc = FormatDesc($desc);
		my ( $d, $c ) = ( $amt, $emptyFields{amount} );
		( $c, $d ) = ( $d, $c ) if $isC;
		push @items, ( SortItem2( $sort, $date, $chk, $d, $c, $desc ) );
	}
	if ( !@items ) {
		$Session->{current}{message} = "No recurring transactions found." if !$PDF;
		return;
	}
	my $closePDF = FALSE;
	if ( !$PDF ) {
		$PDF = PDFOpen( "$ReportsDir/Recurring-$accountname", PORTRAIT, $header, $ColumnHeader );
		$closePDF = TRUE;
	} else {
		PDFHeaders( $PDF, $header, $ColumnHeader );
	}
	if ( !ref($PDF) ) {
		$Session->{current}{error} = $PDF;
		return;
	}
	PDFPrintLine($PDF);
	foreach my $item ( sort @items ) {
		$item =~ s" (\d+)$"";
		$item =~ s" 0 "   "g;
		PDFPrintLine( $PDF, $item );
	}
	PDFPrintLine($PDF);
	PDFPrintLine($PDF);
	SetAndClosePDF($PDF) if $closePDF;
	$Session->{current}{message} = "$header report created.";
}

# ---------------------------------------------------------------------------------------------
sub PrintFullReport {
	my $account = $Session->{current}{account};
	return if !$account;
	my $PDF = PDFOpen( "$ReportsDir/FullReport-$Accounts{$account}{accountname}", PORTRAIT, '', '' );
	PrintUncleared( FALSE, $PDF );
	PrintRecurring($PDF);
	PrintStatement( $Accounts{$account}{currentstatement}, $PDF );
	my %showStatements;
	foreach my $statement ( ( tied %Statements )->select_where("account = '$account'") ) {
		next if $statement eq $Accounts{$account}{currentstatement} or $statement eq $Accounts{$account}{recurringstatement};
		$showStatements{ $Statements{$statement}{statementdate} } = $statement;
	}
	foreach my $showStatement ( reverse sort keys %showStatements ) {
		PrintStatement( $showStatements{$showStatement}, $PDF );
	}
	SetAndClosePDF($PDF);
}

# ---------------------------------------------------------------------------------------------
sub PrintCrDrBalance {
	my ( $accounttype, $balance ) = @_;
	my ( $d, $c ) = ( FormatAmount($balance), $emptyFields{amount} );
	( $c, $d ) = ( $d, $c ) if $accounttype eq 'c';
	return ( $d, $c );
}

# ---------------------------------------------------------------------------------------------
sub PrintBalanceReport {
	my $PDF = PDFOpen( "$ReportsDir/BalanceReport", PORTRAIT, 'Balance Report', '' );
	my %showAccounts;
	foreach my $account ( keys %Accounts ) {
		next if !$Accounts{$account}{archived} ne !$Session->{current}{showarchived};
		$showAccounts{ $Accounts{$account}{accounttype} }{ $Accounts{$account}{accountname} } = $account;
	}
	foreach my $accounttype ( sort keys %showAccounts ) {
		PDFPrintLine($PDF);
		PDFPrintLine( $PDF, $AccountTypes{$accounttype}{label} . " Accounts", 'C', TRUE );
		PDFPrintLine($PDF);
		PDFPrintLine( $PDF, sprintf( "%-30s %15s %15s", 'Account', 'Cleared', 'Balance' ), 'C', TRUE );
		my $ClearedTotal = 0;
		my $BalanceTotal = 0;
		foreach my $AccountName ( sort { uc($a) cmp uc($b) } keys %{ $showAccounts{$accounttype} } ) {
			my $account = $showAccounts{$accounttype}{$AccountName};
			my $cleared = Amount( $Accounts{$account}{clearedbalance} );
			my $balance = Amount( $Accounts{$account}{balance} );
			$ClearedTotal += $cleared;
			$BalanceTotal += $balance;
			PDFPrintLine( $PDF, sprintf( "%-30s %15.2f %15.2f", $AccountName, $cleared, $balance ), 'C' );
		}
		PDFPrintLine( $PDF, '-' x 30 . ' ' . '-' x 15 . ' ' . '-' x 15, 'C', TRUE );
		PDFPrintLine( $PDF, sprintf( "%-30s %15.2f %15.2f", 'TOTAL', $ClearedTotal, $BalanceTotal ), 'C', TRUE );
		PDFPrintLine($PDF);
		PDFDrawLine($PDF);
	}
	SetAndClosePDF($PDF);
	$Session->{current}{message} = "Balance Report created.";
}

# ---------------------------------------------------------------------------------------------
sub PrintCreditUtilization {
	my $PDF = PDFOpen( "$ReportsDir/CreditUtilization", LANDSCAPE, 'Credit Utilization Report', '' );
	PDFFontSize( $PDF, 9 );
	my %showAccounts;
	my %data;
	my %totals;
	foreach my $account ( keys %Accounts ) {
		next if $Accounts{$account}{accounttype} ne 'c';
		next if !$Accounts{$account}{archived} ne !$Session->{current}{showarchived};
		next if !$Accounts{$account}{creditlimit} or !$Accounts{$account}{interestrate};
		$showAccounts{ $Accounts{$account}{accountname} } = $account;
		$data{$account}{balance}                          = Amount( $Accounts{$account}{balance} );
		$data{$account}{limit}                            = Amount( $Accounts{$account}{creditlimit} );
		$data{$account}{rate}                             = Amount( $Accounts{$account}{interestrate} );
		$totals{balance} += $data{$account}{balance};
		$totals{limit}   += $data{$account}{limit};
	}
	foreach my $AccountName ( sort { uc($a) cmp uc($b) } keys %showAccounts ) {
		my $account = $showAccounts{$AccountName};
		$data{$account}{util}     = sprintf "%.2f", ( $data{$account}{balance} / $data{$account}{limit} * 100 );
		$data{$account}{interest} = sprintf "%.2f", ( $data{$account}{balance} * $data{$account}{rate} / 100.0 );
		$totals{interest} += $data{$account}{interest};
	}
	my %totalscore;
	my $nonZero = 0;
	foreach my $AccountName ( sort { uc($a) cmp uc($b) } keys %showAccounts ) {
		my $account = $showAccounts{$AccountName};
		$data{$account}{debtpct} = sprintf "%.2f", ( $data{$account}{balance} / $totals{balance} * 100 );
		if ( $data{$account}{balance} == 0 ) {
			$data{$account}{rscore} = $data{$account}{escore} = $data{$account}{pscore} = $data{$account}{cscore} = 0;
		} else {
			$nonZero++;
			$data{$account}{rscore} = $data{$account}{rate};
			$data{$account}{escore} = $data{$account}{util} * $data{$account}{rate};
			$data{$account}{pscore} = $data{$account}{debtpct} != 0 ? $data{$account}{rate} / $data{$account}{debtpct} : 0;
			$data{$account}{cscore} = $data{$account}{escore} * $data{$account}{pscore};
			foreach ( 'rscore', 'escore', 'pscore', 'cscore' ) { $totalscore{$_} += $data{$account}{$_}; }
		}
		$data{$account}{adjlimit} = $data{$account}{limit} / $data{$account}{rate};
		$totals{adjlimit} += $data{$account}{adjlimit};
	}
	$totals{utilpct}    = $totals{balance} / $totals{limit};
	$totals{adjutilpct} = $totals{balance} / $totals{adjlimit};
	foreach my $AccountName ( sort { uc($a) cmp uc($b) } keys %showAccounts ) {
		my $account = $showAccounts{$AccountName};
		$data{$account}{adjbalance}  = sprintf "%.2f", $totals{utilpct} * $data{$account}{limit};
		$data{$account}{adjbalancew} = sprintf "%.2f", $totals{adjutilpct} * $data{$account}{adjlimit};
		$data{$account}{adjust}      = sprintf "%.2f", $data{$account}{adjbalance} - $data{$account}{balance};
		$data{$account}{adjustw}     = sprintf "%.2f", $data{$account}{adjbalancew} - $data{$account}{balance};
		$data{$account}{adjutil}      = sprintf "%.2f", ( $data{$account}{adjbalance} / $data{$account}{limit} * 100 );
		$data{$account}{adjutilw}     = sprintf "%.2f", ( $data{$account}{adjbalancew} / $data{$account}{limit} * 100 );
		$data{$account}{adjinterest}  = sprintf "%.2f", ( $data{$account}{adjbalance} * $data{$account}{rate} / 100.0 );
		$data{$account}{adjinterestw} = sprintf "%.2f", ( $data{$account}{adjbalancew} * $data{$account}{rate} / 100.0 );
		$totals{adjust} += $data{$account}{adjust};
	}
	my %sort;
	foreach my $AccountName ( sort { uc($a) cmp uc($b) } keys %showAccounts ) {
		my $account = $showAccounts{$AccountName};
		foreach my $score ( 'rscore', 'escore', 'pscore', 'cscore' ) {
			$data{$account}{$score} = $data{$account}{$score} / $totalscore{$score} * 100;
			my $order = sprintf "%.5d", $data{$account}{$score} * 100;
			$sort{$score}{$order}{$AccountName} = 1;
		}
	}
	my @fields   = ( 'balance', 'limit', 'util',     'debtpct', 'rate', 'interest', 'adjust', 'adjustw', 'adjutil', 'adjutilw', 'adjinterest', 'adjinterestw', 'rscore', 'escore', 'pscore', 'cscore' );
	my @reportr  = ( 'balance', 'rate',  'interest', 'rscore' );
	my @reporte  = ( 'balance', 'limit', 'util',     'rate', 'adjust', 'adjutil', 'adjinterest', 'escore' );
	my @reportew = ( 'balance', 'limit', 'util',     'rate', 'adjustw', 'adjutilw', 'adjinterestw', 'escore' );
	my @reportp  = ( 'balance', 'limit', 'debtpct',  'rate', 'interest', 'pscore' );
	my @reportc  = ( 'balance', 'limit', 'util',     'debtpct', 'rate', 'cscore' );
	my %reports = ( r => \@reportr, e => \@reporte, ew => \@reportew, p => \@reportp, c => \@reportc, );
	my %order = ( r => 'rscore', e => 'escore', ew => 'escore', p => 'pscore', c => 'cscore' );
	my %title = ( r => 'Rate Analysis', e => 'Credit Equalization Analysis', ew => 'Rate Weighted Credit Equalization Analysis', p => 'Payoff Analysis', c => 'Combined Analysis' );
	my %labels = (
		balance      => 'Current Balance',
		limit        => 'Credit Limit',
		util         => 'Credit Utilization',
		debtpct      => 'Debt Percentage',
		rate         => 'Interest Rate',
		interest     => 'Annual Interest',
		adjust       => 'Balance Adjustment',
		adjustw      => 'Balance Adjustment',
		adjutil      => 'Adjusted Utilization',
		adjutilw     => 'Adjusted Utilization',
		adjinterest  => 'Adjusted Interest',
		adjinterestw => 'Adjusted Interest',
		rscore       => 'Rate Score',
		escore       => 'Weighted Score',
		pscore       => 'Payoff Score',
		cscore       => 'Combined Score',
	);
	my %description = (
		r  => "This is a basic analysis based on the $labels{rate}, showing higher rates first, and scoring rates from high to low.\nThis report is used to show which accounts cost the most per month.",
		e  => "This is an analysis with a score combining $labels{rate} and $labels{util}, showing a $labels{adjust},\n  not weighted by $labels{rate}.\nThis report can be used to even out Credit Utilization across accounts and improve credit scores.",
		ew => "This is an analysis with a score combining $labels{rate} and $labels{util}, showing a $labels{adjust},\n  weighted by $labels{rate}.\nThis report can be used to even out Credit Utilization across accounts and improve credit scores.",
		p  => "This is an analysis combining $labels{rate} and $labels{debtpct}.\nThis report can be used for paying off smaller, more expensive accounts first.",
		c  => "This analysis is a combination of the Credit Equalization Analysis and Payoff Analysis.",
	);
	my %doTotals = (
		balance      => TRUE,
		limit        => TRUE,
		util         => '',
		debtpct      => TRUE,
		rate         => '',
		interest     => TRUE,
		rscore       => TRUE,
		escore       => TRUE,
		pscore       => TRUE,
		adjust       => TRUE,
		adjustw      => TRUE,
		adjutil      => '',
		adjutilw     => '',
		adjinterest  => TRUE,
		adjinterestw => TRUE,
	);
	my @labels1;
	my @labels2;
	my ( %labels1, %labels2 );

	foreach my $fld (@fields) {
		my $label = $labels{$fld};
		my ( $label1, $label2 ) = split m" +", $label;
		( $label1, $label2 ) = ( '', $label1 ) if !$label2;
		$labels1{$fld} = $label1;
		$labels2{$fld} = $label2;
	}
	my @seps;
	my %total;
	my $sep30   = '-' x 30;
	my $sep11   = '-' x 11;
	my $formatn = '%-30s';
	my $formatd = ' %12.2f';
	my $formatl = ' %12s';

	foreach my $report ( 'r', 'e', 'ew', 'p', 'c' ) {
		my @rfields = @{ $reports{$report} };
		my $format1 = my $format2 = $formatn;
		foreach my $fld (@rfields) { $format1 .= $formatd; $format2 .= $formatl; $total{$fld} = 0; push @seps, ($sep11); }
		PDFPrintHeader($PDF);
		PDFPrintLine($PDF);
		PDFPrintLine( $PDF, $title{$report}, 'c', TRUE );
		PDFPrintLine($PDF);
		PDFPrintLine($PDF);
		my $hdr1 = sprintf $formatn, '';
		my $hdr2 = sprintf $formatn, 'Account';
		foreach my $fld ( @{ $reports{$report} } ) { $hdr1 .= sprintf $formatl, $labels1{$fld}; $hdr2 .= sprintf $formatl, $labels2{$fld}; }    # format1 .= $formatd; $format2 .= $formatl; $total{$fld} = 0; push @seps, ($sep10); }
		PDFPrintLine( $PDF, $hdr1, '', TRUE );
		PDFPrintLine( $PDF, $hdr2, '', TRUE );
		my $count    = 0;
		my $sortkey1 = $order{$report};

		foreach my $score ( reverse sort keys %{ $sort{$sortkey1} } ) {
			foreach my $AccountName ( sort { uc($a) cmp uc($b) } keys %{ $sort{$sortkey1}{$score} } ) {
				$count++;
				my $account = $showAccounts{$AccountName};
				my @values;
				foreach my $fld (@rfields) {
					$total{$fld} += $data{$account}{$fld};
					push @values, ( $data{$account}{$fld} );
				}
				PDFPrintLine( $PDF, sprintf( $format1, $AccountName, @values ) );
			}
		}
		PDFPrintLine( $PDF, sprintf( $format2, $sep30, @seps ) );
		my $totalLine = sprintf $formatn, 'TOTAL';
		foreach my $fld (@rfields) { $totalLine .= $doTotals{$fld} ? sprintf $formatd, $total{$fld} : sprintf $formatl, ''; }
		PDFPrintLine( $PDF, $totalLine, '', TRUE );
		$totalLine = sprintf $formatn, 'AVERAGE';
		foreach my $fld (@rfields) { $totalLine .= $doTotals{$fld} ? sprintf $formatl, '' : sprintf $formatd, $total{$fld} / $count; }
		PDFPrintLine( $PDF, $totalLine, '', TRUE );
		PDFPrintLine($PDF);
		PDFDrawLine($PDF);

		if ( $report eq 'r' ) {
			PDFPrintLine($PDF);
			PDFPrintLine( $PDF, sprintf( "Average interest rate is: %.2f%10s(This is Total Interest divided by Total Balance.)", ( $totals{interest} / $totals{balance} * 100 ), '' ) );
			PDFPrintLine($PDF);
			PDFDrawLine($PDF);
		}
		foreach my $line ( '', split m"\n", $description{$report} ) {
			PDFPrintLine( $PDF, $line );
		}
	}
	SetAndClosePDF($PDF);
}

# ---------------------------------------------------------------------------------------------
sub SetAndClosePDF {
	my ($PDF) = @_;
	PDFClose($PDF);
	$DownloadFile = $$PDF{Filename};
}

# ---------------------------------------------------------------------------------------------
# End
1;
__END__
