package Server;
use strict;
require Exporter;
use vars qw (@ISA @EXPORT);
@ISA    = qw(Exporter);                                      # Inherit from Exporter
@EXPORT = qw( SetPort OpenSocket StartBrowser RunServer );
use strict;
use warnings;
use IO::File;
use File::Basename;
use IO::Socket;
use Sys::Hostname;
use Common;
use Multiuser;
use Data;
use HTML;
use Authenticate;
use Configure;
use Pages;
use Process;

# ---------------------------------------------------------------------------------------------
# set port
sub SetPort {
	my $port = $properties{'port'};
	if ( !PortIsAvailable($port) ) {
		die "Port $port is not available!\n" if !$properties{'dynamicport'};
		until ( PortIsAvailable($port) ) { $port++; }
	}
	$properties{'port'} = $port;
	$properties{'url'}  = 'http://' . hostname() . ':' . $properties{port};
}

# ---------------------------------------------------------------------------------------------
sub OpenSocket {
	our $socket = new IO::Socket::INET( LocalPort => $properties{port}, Listen => SOMAXCONN, Proto => 'tcp', Reuse => 1, Type => SOCK_STREAM );
}

# ---------------------------------------------------------------------------------------------
# RunServer
sub RunServer {
	while ( my $request = our $socket->accept() ) {
		$request->autoflush(1);
		binmode $request;
		HandleRequest($request);
		close $request;
		exit(0) if $Shutdown;
	}
}

# ---------------------------------------------------------------------------------------------
# Handle Request
sub HandleRequest {
	my ($fh) = @_;
	$clientIP = $fh->peerhost;
	my $contentLength = 0;
	my $fhin          = "";
	my $fhout;
	vec( $fhin, fileno($fh), 1 ) = 1;
	my ( $nfound, $timeleft ) = select( $fhout = $fhin, undef, undef, 0.01 );
	return if !$nfound;
	my $firstLine = <$fh>;
	return if !$firstLine;    # socket connection with no request
	$firstLine =~ s"\r?\n$"";
	my ( $method, $object, $protocol ) = split m" +", $firstLine;
	$method = uc $method if $method;
	return if $method ne "POST" and $method ne "GET";
	my $requestLine = TRUE;
	Debug( '', $firstLine );
	$UserAgent = "";

	while ($requestLine) {
		$requestLine = <$fh>;
		$requestLine =~ s"\r?\n$"" if $requestLine;
		$contentLength = $1 if $requestLine =~ m"content-length:\D+(\d+)"i;
		$UserAgent     = $1 if $requestLine =~ m"^User-Agent: (.*)";
	}
	$UserAgent = "Chrome"  if $UserAgent =~ m"chrome"i;
	$UserAgent = "Firefox" if $UserAgent =~ m"firefox"i;
	$UserAgent = "IE$1"    if $UserAgent =~ m"MSIE (\d+)\."i;
	$object =~ s"^/"";
	my $content = "";
	if ($contentLength) {
		read( $fh, $content, $contentLength );
		$content =~ s/\+/ /g;
	}
	select $fh;
	$| = 1;
	PageValues($content);
	return Configure($content) if $object eq 'configure';
	return Users($content)     if $object eq 'users';
	if ( $method eq 'GET' and $object and $object !~ m"^\d+(/.*)?$" ) {
		return ReturnFile($object);
	}
	my $urlcontent = "";
	$urlcontent = $1 if $object =~ s"^(\d+)/(.*)$"$1";
	Debug("object '$object'");
	return HTTPForbidden() if $object and $object ne 'login' and $object !~ m"^\d*$";
	SetNow();
	CheckBackupSchedule();
	my $view = $input{view};
	$view ||= '';

	if ( $view eq 'logout' or $view eq 'quit' or $view eq 'shutdown' ) {
		$Shutdown = TRUE if $view eq 'shutdown' or $properties{shutdownonquit};
		SaveSession();
		return HTMLCloseWindow();
	}
	return if !CheckAuthentication($object);
	$DownloadContent = $DownloadFile = $DownloadContentType = $DownloadFilename = '';
	StartDBTransaction();
	ProcessPage($urlcontent);
	FinishDBTransaction();
	return Download() if $DownloadContent or $DownloadFile;
	ShowPage();
}

# ---------------------------------------------------------------------------------------------
sub ReturnFile {
	my ($File) = @_;
	return HTTPForbidden() if $File !~ m"\.(js|css|cur|ico)$";
	return HTTPNotFound($File) if !-f $File;
	our %fileContent;
	if ( !$fileContent{$File} ) {
		open my $FILE, "<", $File;
		my @Content = <$FILE>;
		close $FILE;
		$fileContent{$File} = join( '', @Content );
	}
	my $type = "text/plain";
	$type = 'application/javascript' if $File =~ m"\.js$";
	$type = 'text/css'               if $File =~ m"\.css$";
	$type = 'text/plain'             if $File =~ m"\.cur$";
	$type = 'image/x-icon'           if $File =~ m"\.ico$";
	my $body   = $fileContent{$File};
	my $length = length($body);
	print "HTTP/1.1 200 OK\nContent-Type: $type\nCache-Control: no-cache\nContent-Length: $length\n\n$body";
	Debug("returned '$File'");
}

# ---------------------------------------------------------------------------------------------
sub Download {
	$DownloadContentType = $1 if !$DownloadContentType and $DownloadFilename =~ m"\.(...)$";
	$DownloadContentType = $1 if !$DownloadContentType and $DownloadFile =~ m"\.(...)$";
	if ($DownloadFile) {
		open my $DOWNLOADFILE, "<", $DownloadFile;
		binmode $DOWNLOADFILE;
		read( $DOWNLOADFILE, $DownloadContent, -s $DownloadFile );
		close $DOWNLOADFILE;
	}
	my $attachment = $DownloadFilename ? "Content-Disposition: attachment; filename=$DownloadFilename\n" : "";
	my $length = length($DownloadContent);
	print "HTTP/1.1 200 OK\nContent-Type: application/$DownloadContentType\nCache-Control: no-cache\nContent-length: $length\n$attachment\n", $DownloadContent;
}

# ---------------------------------------------------------------------------------------------
# set browsers
sub SetBrowsers {
	my @choices = ();
	foreach my $browser ( 'iexplore', 'firefox', 'chrome' ) {
		my @browsers = glob("c:/Program*/*/${browser}.exe");
		my ($browserexe) = grep m"x86", @browsers;    # look for 32-bit first
		($browserexe) = @browsers if !$browserexe;    # take first if any
		if ( !$browserexe ) {
			my @dirs = ( $ENV{homedrive} . $ENV{homepath} );
			while ( my $dir = shift @dirs ) {
				($browserexe) = glob("$dir/${browser}.exe");
				last if $browserexe and -f $browserexe;
				foreach my $file ( glob("$dir/*") ) {
					next if $file =~ m"/\.\.?$";
					next if !-d $file;
					push @dirs, ($file);
				}
			}
		}
		if ($browserexe) {
			push @choices, ($browser);
		} else {
			$browserexe = 'No ${browser}.exe found!';
		}
		$properties{$browser} = $browserexe;
	}
	$properties{browser} = 'default - select ' . join( ' or ', @choices ) . ' to override.';
}

# ---------------------------------------------------------------------------------------------
# StartBrowser
sub StartBrowser {
	SetBrowsers() if !$properties{browser};
	my $Browser = $properties{browser};
	$Browser =~ s" .*"";
	if ( $Browser =~ m"^default"i ) {
		system("start $properties{url}");
		return;
	}
	$Browser = $properties{$Browser} if $Browser and $properties{$Browser};
	SetBrowsers() if !-e $Browser;
	my $BrowserOptions = "";                                                # for future use, such as kiosk mode
	my $BrowserCommand = "\"$Browser\" $BrowserOptions $properties{url}";
	system("start \"\" $BrowserCommand");
}

# ---------------------------------------------------------------------------------------------
sub CheckBackupSchedule {                                                   #backupschedule is frequency in hours
	return if !$properties{backupschedule};
	return if $Now{seconds} - $LastBackupTime < ( $properties{backupschedule} * 3600 );
	CloseDatabase();
	OpenDatabase();                                                         #backups are done in OpenDatabase
}

# ---------------------------------------------------------------------------------------------
sub Debug { }                                                               # my (@debug) = @_; print STDERR join( "\n", @debug ), "\n"; }

# ---------------------------------------------------------------------------------------------
# End
1;
